#! /usr/bin/env python

"""
Module with sky subtraction function.
"""

from __future__ import division

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['cube_subtract_sky_pca']

import numpy as np
from ..var import prepare_matrix


def cube_subtract_sky_pca(sci_cube, sky_cube, mask, ref_cube=None, ncomp=2):
    """ PCA based sky subtraction.

    Parameters
    ----------
    sci_cube : array_like
        3d array of science frames.
    sky_cube : array_like
        3d array of sky frames.
    mask : array_like
        Mask indicating the region for the analysis. Can be created with the
        function vip_hci.var.create_ringed_spider_mask.
    ref_cube : array_like or None
        Reference cube.
    ncomp : int
        Sets the number of PCs you want to use in the sky subtraction.

    Returns
    -------
    Sky subtracted cube.

    """
    from ..pca import svd_wrapper

    if sci_cube.shape[1] != sky_cube.shape[1] or sci_cube.shape[2] != \
            sky_cube.shape[2]:
        raise TypeError('Science and Sky frames sizes do not match')
    if ref_cube is not None:
        if sci_cube.shape[1] != ref_cube.shape[1] or sci_cube.shape[2] != \
                ref_cube.shape[2]:
            raise TypeError('Science and Reference frames sizes do not match')

    # Getting the EVs from the sky cube
    Msky = prepare_matrix(sky_cube, scaling=None, verbose=False)
    sky_pcs = svd_wrapper(Msky, 'lapack', sky_cube.shape[0], False,
                                  False)
    sky_pcs_cube = sky_pcs.reshape(sky_cube.shape[0], sky_cube.shape[1],
                                   sky_cube.shape[1])

    # Masking the science cube
    sci_cube_masked = np.zeros_like(sci_cube)
    ind_masked = np.where(mask == 0)
    for i in range(sci_cube.shape[0]):
        masked_image = np.copy(sci_cube[i])
        masked_image[ind_masked] = 0
        sci_cube_masked[i] = masked_image
    Msci_masked = prepare_matrix(sci_cube_masked, scaling=None,
                                         verbose=False)

    # Masking the PCs learned from the skies
    sky_pcs_cube_masked = np.zeros_like(sky_pcs_cube)
    for i in range(sky_pcs_cube.shape[0]):
        masked_image = np.copy(sky_pcs_cube[i])
        masked_image[ind_masked] = 0
        sky_pcs_cube_masked[i] = masked_image

    # Project the masked frames onto the sky PCs to get the coefficients
    transf_sci = np.zeros((sky_cube.shape[0], Msci_masked.shape[0]))
    for i in range(Msci_masked.shape[0]):
        transf_sci[:, i] = np.inner(sky_pcs, Msci_masked[i].T)

    Msky_pcs_masked = prepare_matrix(sky_pcs_cube_masked, scaling=None,
                                     verbose=False)
    mat_inv = np.linalg.inv(np.dot(Msky_pcs_masked, Msky_pcs_masked.T))
    transf_sci_scaled = np.dot(mat_inv, transf_sci)

    # Obtaining the optimized sky and subtraction
    sci_cube_skysub = np.zeros_like(sci_cube)
    for i in range(Msci_masked.shape[0]):
        sky_opt = np.array([np.sum(
            transf_sci_scaled[j, i] * sky_pcs_cube[j] for j in range(ncomp))])
        sci_cube_skysub[i] = sci_cube[i] - sky_opt

    # Processing the reference cube (if any)
    if ref_cube is not None:
        ref_cube_masked = np.zeros_like(ref_cube)
        for i in range(ref_cube.shape[0]):
            masked_image = np.copy(ref_cube[i])
            masked_image[ind_masked] = 0
            ref_cube_masked[i] = masked_image
        Mref_masked = prepare_matrix(ref_cube_masked, scaling=None,
                                             verbose=False)
        transf_ref = np.zeros((sky_cube.shape[0], Mref_masked.shape[0]))
        for i in range(Mref_masked.shape[0]):
            transf_ref[:, i] = np.inner(sky_pcs, Mref_masked[i].T)

        transf_ref_scaled = np.dot(mat_inv, transf_ref)

        ref_cube_skysub = np.zeros_like(ref_cube)
        for i in range(Mref_masked.shape[0]):
            sky_opt = np.array([np.sum(
                transf_ref_scaled[j, i] * sky_pcs_cube[j] for j in range(ncomp))])
            ref_cube_skysub[i] = ref_cube[i] - sky_opt

        return sci_cube_skysub, ref_cube_skysub
    else:
        return sci_cube_skysub



#! /usr/bin/env python

"""
Module with post-processing related functions called from within the NFC
algorithm.
"""

__all__ = ['cube_planet_free']

import numpy as np
from ..phot import cube_inject_companions
import math
from matplotlib.pyplot import plot, xlim, ylim, hold, axes, gca, show


def cube_planet_free(planet_parameter, cube, angs, psfn, plsc, imlib='opencv',
                     interpolation='lanczos4'):
    """
    Return a cube in which we have injected negative fake companion at the
    position/flux given by planet_parameter.

    Parameters
    ----------
    planet_parameter: numpy.array or list
        The (r, theta, flux) for all known companions.
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array.
    psfsn: numpy.array
        The scaled psf expressed as a numpy.array.
    plsc: float
        The platescale, in arcsec per pixel.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.

    Returns
    -------
    cpf : numpy.array
        The cube with negative companions injected at the position given in
        planet_parameter.

    """
    cpf = np.zeros_like(cube)

    planet_parameter = np.array(planet_parameter)

    for i in range(planet_parameter.shape[0]):
        if i == 0:
            cube_temp = cube
        else:
            cube_temp = cpf

        cpf = cube_inject_companions(cube_temp, psfn, angs,
                                     flevel=-planet_parameter[i, 2],
                                     plsc=plsc, rad_dists=[planet_parameter[i, 0]],
                                     n_branches=1, theta=planet_parameter[i, 1],
                                     imlib=imlib, interpolation=interpolation,
                                     verbose=False)
    return cpf


def radial_to_eq(r=1, t=0, rError=0, tError=0, display=False):
    """ 
    Convert the position given in (r,t) into \delta RA and \delta DEC, as 
    well as the corresponding uncertainties. 
    t = 0 deg (resp. 90 deg) points toward North (resp. East).   

    Parameters
    ----------
    r: float
        The radial coordinate.
    t: float
        The angular coordinate.
    rError: float
        The error bar related to r.
    tError: float
        The error bar related to t.
    display: boolean, optional
        If True, a figure illustrating the error ellipse is displayed.
        
    Returns
    -------
    out : tuple
        ((RA, RA error), (DEC, DEC error))
                              
    """  
    ra = (r * np.sin(math.radians(t)))
    dec = (r * np.cos(math.radians(t)))   
    u, v = (ra, dec)
    
    nu = np.mod(np.pi/2.-math.radians(t), 2*np.pi)
    a, b = (rError,r*np.sin(math.radians(tError)))

    beta = np.linspace(0,2*np.pi,5000)
    x, y = (u + (a * np.cos(beta) * np.cos(nu) - b * np.sin(beta) * np.sin(nu)),
            v + (b * np.sin(beta) * np.cos(nu) + a * np.cos(beta) * np.sin(nu)))
    
    raErrorInf = u - np.amin(x)
    raErrorSup = np.amax(x) - u
    decErrorInf = v - np.amin(y)
    decErrorSup = np.amax(y) - v        

    if display:        
        hold(True)
        plot(u,v,'ks',x,y,'r')        
        plot((r+rError) * np.cos(nu), (r+rError) * np.sin(nu),'ob',
             (r-rError) * np.cos(nu), (r-rError) * np.sin(nu),'ob')
        plot(r * np.cos(nu+math.radians(tError)), 
             r*np.sin(nu+math.radians(tError)),'ok')
        plot(r * np.cos(nu-math.radians(tError)), 
             r*np.sin(nu-math.radians(tError)),'ok')
        plot(0,0,'og',np.cos(np.linspace(0,2*np.pi,10000)) * r, 
             np.sin(np.linspace(0,2*np.pi,10000)) * r,'y')
        plot([0,r*np.cos(nu+math.radians(tError*0))],
             [0,r*np.sin(nu+math.radians(tError*0))],'k')
        axes().set_aspect('equal')
        lim = np.amax([a,b]) * 2.
        xlim([ra-lim,ra+lim])
        ylim([dec-lim,dec+lim])
        gca().invert_xaxis()
        show()
        
    return ((ra,np.mean([raErrorInf,raErrorSup])),
            (dec,np.mean([decErrorInf,decErrorSup])))    
    
    
def cart_to_polar(y, x, ceny=0, cenx=0):
    """
    Convert cartesian into polar coordinates (r,theta) with 
    respect to a given center (cenx,ceny).
    
    Parameters
    ----------
    x,y: float
        The cartesian coordinates.
        
    Returns
    -------
    
    out : tuple
        The polar coordinates (r,theta) with respect to the (cenx,ceny). 
        Note that theta is given in degrees.
        
    """
    r = np.sqrt((y-ceny)**2 + (x-cenx)**2)
    theta = np.degrees(np.arctan2(y-ceny, x-cenx)) 
    
    return (r,np.mod(theta,360))   


def polar_to_cart(r, theta, ceny=0, cenx=0):
    """
    Convert polar coordinates with respect to the center (cenx,ceny) into 
    cartesian coordinates (x,y) with respect to the bottom left corner of the 
    image..
    
    Parameters
    ----------
    r,theta: float
        The polar coordinates.
        
    Returns
    -------
    
    out : tuple
        The cartesian coordinates (x,y) with respect to the bottom left corner 
        of the image..
        
    """
    x = r*np.cos(np.deg2rad(theta)) + cenx
    y = r*np.sin(np.deg2rad(theta)) + ceny 
    
    return (x,y)    


def ds9index_to_polar(y, x, ceny=0, cenx=0):
    """
    Convert pixel index read on image displayed with DS9 into polar coordinates 
    (r,theta) with respect to a given center (cenx,ceny).
    
    Note that ds9 index (x,y) = Python matrix index (y,x). Furthermore, when an 
    image M is displayed with DS9, the coordinates of the center of the pixel 
    associated with M[0,0] is (1,1). Then, there is a shift of (0.5, 0.5) of the
    center of the coordinate system. As a conclusion, when you read (x_ds9, y_ds9)
    on a image displayed with DS9, the corresponding position is (y-0.5, x-0.5)
    and the associated pixel value is M(np.floor(y)-1,np.floor(x)-1).
    
    Parameters
    ----------
    x,y: float
        The pixel index in DS9
        
    Returns
    -------
    
    out : tuple
        The polar coordinates (r,theta) with respect to the (cenx,ceny). 
        Note that theta is given in degrees.
        
    """
    r = np.sqrt((y-0.5-ceny)**2 + (x-0.5-cenx)**2)
    theta = np.degrees(np.arctan2(y-0.5-ceny, x-0.5-cenx))
    
    return (r,np.mod(theta,360)) 
    
    
def polar_to_ds9index(r, theta, ceny=0, cenx=0):
    """
    Convert position (r,theta) in an image with respect to a given center 
    (cenx,ceny) into position in the image displayed with DS9.
    
    Note that ds9 index (x,y) = Python matrix index (y,x). Furthermore, when an 
    image M is displayed with DS9, the coordinates of the center of the pixel 
    associated with M[0,0] is (1,1). Then, there is a shift of (0.5, 0.5) of the
    center of the coordinate system. As a conclusion, when you read (x_ds9, y_ds9)
    on a image displayed with DS9, the corresponding position is (y-0.5, x-0.5)
    and the associated pixel value is M(np.floor(y)-1,np.floor(x)-1).
    
    Parameters
    ----------
    x,y: float
        The pixel index in DS9
        
    Returns
    -------
    
    out : tuple
        The polar coordinates (r,theta) with respect to the (cenx,ceny). 
        Note that theta is given in degrees.
        
    """
    x_ds9 = r*np.cos(np.deg2rad(theta)) + 0.5 + cenx
    y_ds9 = r*np.sin(np.deg2rad(theta)) + 0.5 + ceny 
    
    return (x_ds9,y_ds9)    


#! /usr/bin/env python

"""
Module containing functions for cubes frame registration.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez, V. Christiaens @ ULg/UChile, G. Ruane'
__all__ = ['frame_shift',
           'frame_center_radon',
           'frame_center_satspots',
           'cube_recenter_satspots',
           'cube_recenter_radon',
           'cube_recenter_dft_upsampling',
           'cube_recenter_2dfit',
           'cube_recenter_via_speckles']

import numpy as np
import warnings
import itertools as itt
import pyprind

try:
    import cv2
    no_opencv = False
except ImportError:
    msg = "Opencv python bindings are missing."
    warnings.warn(msg, ImportWarning)
    no_opencv = True

from scipy.ndimage import fourier_shift
from scipy.ndimage import shift
from skimage.transform import radon
from skimage.feature import register_translation
from multiprocessing import Pool, cpu_count
from matplotlib import pyplot as plt
from . import frame_crop
from ..conf import time_ini, timing
from ..conf.utils_conf import eval_func_tuple as EFT
from ..var import (get_square, frame_center, get_annulus, pp_subplots,
                   fit_2dmoffat, fit_2dgaussian, frame_filter_gaussian2d,
                   frame_filter_lowpass)
from ..preproc import cube_crop_frames


def frame_shift(array, shift_y, shift_x, imlib='opencv',
                interpolation='lanczos4'):
    """ Shifts a 2D array by shift_y, shift_x. Boundaries are filled with zeros.

    Parameters
    ----------
    array : array_like
        Input 2d array.
    shift_y, shift_x: float
        Shifts in x and y directions.
    imlib : {'opencv', 'ndimage-fourier', 'ndimage-interp'}, string optional
        Library or method used for performing the image shift.
        'ndimage-fourier', does a fourier shift operation and preserves better
        the pixel values (therefore the flux and photometry). Interpolation
        based shift ('opencv' and 'ndimage-interp') is faster than the fourier
        shift. 'opencv' is recommended when speed is critical.
    interpolation : {'bicubic', 'bilinear', 'nearneig'}, optional
        Only used in case of imlib is set to 'opencv' or 'ndimage-interp', where
        the images are shifted via interpolation.
        For 'ndimage-interp' library: 'nearneig', bilinear', 'bicuadratic',
        'bicubic', 'biquartic', 'biquintic'. The 'nearneig' interpolation is
        the fastest and the 'biquintic' the slowest. The 'nearneig' is the
        poorer option for interpolation of noisy astronomical images.
        For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
        The 'nearneig' interpolation is the fastest and the 'lanczos4' the
        slowest and accurate. 'lanczos4' is the default.
    
    Returns
    -------
    array_shifted : array_like
        Shifted 2d array.

    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array')
    
    image = array.copy()

    if imlib == 'ndimage-fourier':
        shift_val = (shift_y, shift_x)
        array_shifted = fourier_shift(np.fft.fftn(image), shift_val)
        array_shifted = np.fft.ifftn(array_shifted)
        array_shifted = array_shifted.real

    elif imlib == 'ndimage-interp':
        if interpolation == 'nearneig':
            order = 0
        elif interpolation == 'bilinear':
            order = 1
        elif interpolation == 'bicuadratic':
            order = 2
        elif interpolation == 'bicubic':
            order = 3
        elif interpolation == 'biquartic':
            order = 4
        elif interpolation == 'biquintic':
            order = 5
        else:
            raise TypeError('Scipy.ndimage interpolation method not recognized.')
        
        array_shifted = shift(image, (shift_y, shift_x), order=order)
    
    elif imlib == 'opencv':
        if no_opencv:
            msg = 'Opencv python bindings cannot be imported. Install opencv or '
            msg += 'set imlib to ndimage-fourier or ndimage-interp'
            raise RuntimeError(msg)

        if interpolation == 'bilinear':
            intp = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            intp= cv2.INTER_CUBIC
        elif interpolation == 'nearneig':
            intp = cv2.INTER_NEAREST
        elif interpolation == 'lanczos4':
            intp = cv2.INTER_LANCZOS4
        else:
            raise TypeError('Opencv interpolation method not recognized.')
        
        image = np.float32(image)
        y, x = image.shape
        M = np.float32([[1,0,shift_x],[0,1,shift_y]])
        array_shifted = cv2.warpAffine(image, M, (x,y), flags=intp)

    else:
        raise ValueError('Image transformation library not recognized.')
    
    return array_shifted


def frame_center_satspots(array, xy, subim_size=19, sigfactor=6, shift=False,
                          imlib='opencv', interpolation='lanczos4',
                          debug=False):
    """ Finds the center of a frame with waffle/satellite spots (e.g. for 
    VLT/SPHERE). The method used to determine the center is by centroiding the
    4 spots via a 2d Gaussian fit and finding the intersection of the 
    lines they create (see Notes). This method is very sensitive to the SNR of 
    the satellite spots, therefore thresholding of the background pixels is 
    performed. If the results are too extreme, the debug parameter will allow to 
    see in depth what is going on with the fit (maybe you'll need to adjust the 
    sigfactor for the background pixels thresholding).  
    
    Parameters
    ----------
    array : array_like, 2d
        Image or frame.
    xy : tuple
        Tuple with coordinates X,Y of the satellite spots in this order:
        upper left, upper right, lower left, lower right. 
    subim_size : int, optional
        Size of subimage where the fitting is done.
    sigfactor : int, optional
        The background pixels will be thresholded before fitting a 2d Gaussian
        to the data using sigma clipped statistics. All values smaller than
        (MEDIAN + sigfactor*STDDEV) will be replaced by small random Gaussian 
        noise. 
    shift : {False, True}, optional 
        If True the image is shifted.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    debug : {False, True}, optional 
        If True debug information is printed and plotted.
    
    Returns
    -------
    shifty, shiftx 
        Shift Y,X to get to the true center.
    If shift is True then the shifted image is returned along with the shift. 
    
    Notes
    -----
    linear system:
    A1 * x + B1 * y = C1
    A2 * x + B2 * y = C2
    
    Cramer's rule - solution can be found in determinants:
    x = Dx/D
    y = Dy/D
    where D is main determinant of the system:  A1 B1
                                                A2 B2
    and Dx and Dy can be found from matrices:  C1 B1
                                               C2 B2
    and  A1 C1
         A2 C2
    C column consequently substitutes the coef. columns of x and y

    L stores our coefs A, B, C of the line equations.
    For D: L1[0] L1[1]   for Dx: L1[2] L1[1]   for Dy: L1[0] L1[2]
           L2[0] L2[1]           L2[2] L2[1]           L2[0] L2[2]
    """
    def line(p1, p2):
        """ produces coefs A, B, C of line equation by 2 points
        """
        A = (p1[1] - p2[1])
        B = (p2[0] - p1[0])
        C = (p1[0]*p2[1] - p2[0]*p1[1])
        return A, B, -C

    def intersection(L1, L2):
        """ finds intersection point (if any) of 2 lines provided 
        by coefs
        """
        D  = L1[0] * L2[1] - L1[1] * L2[0]
        Dx = L1[2] * L2[1] - L1[1] * L2[2]
        Dy = L1[0] * L2[2] - L1[2] * L2[0]
        if D != 0:
            x = Dx / D
            y = Dy / D
            return x,y
        else:
            return False
        
    #---------------------------------------------------------------------------
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array')
    if not len(xy) == 4:
        raise TypeError('Input waffle spot coordinates in wrong format')
    
    cy, cx = frame_center(array)
    
    # Upper left
    si1, y1, x1 = get_square(array, subim_size, xy[0][1], xy[0][0], position=True)
    cent2dgx_1, cent2dgy_1 = fit_2dgaussian(si1, theta=135, crop=False, 
                                            threshold=True, sigfactor=sigfactor, 
                                            debug=debug)
    cent2dgx_1 += x1
    cent2dgy_1 += y1
    # Upper right
    si2, y2, x2 = get_square(array, subim_size, xy[1][1], xy[1][0], position=True)
    cent2dgx_2, cent2dgy_2 = fit_2dgaussian(si2, theta=45, crop=False, 
                                            threshold=True, sigfactor=sigfactor, 
                                            debug=debug)
    cent2dgx_2 += x2
    cent2dgy_2 += y2 
    #  Lower left
    si3, y3, x3 = get_square(array, subim_size, xy[2][1], xy[2][0], position=True)
    cent2dgx_3, cent2dgy_3 = fit_2dgaussian(si3, theta=45, crop=False, 
                                            threshold=True, sigfactor=sigfactor, 
                                            debug=debug)
    cent2dgx_3 += x3
    cent2dgy_3 += y3
    #  Lower right
    si4, y4, x4 = get_square(array, subim_size, xy[3][1], xy[3][0], position=True)
    cent2dgx_4, cent2dgy_4 = fit_2dgaussian(si4, theta=135, crop=False, 
                                            threshold=True, sigfactor=sigfactor, 
                                            debug=debug)
    cent2dgx_4 += x4
    cent2dgy_4 += y4
    
    if debug: 
        pp_subplots(si1, si2, si3, si4, colorb=True)
        print('Centroids X,Y:')
        print(cent2dgx_1, cent2dgy_1)
        print(cent2dgx_2, cent2dgy_2)
        print(cent2dgx_3, cent2dgy_3)
        print(cent2dgx_4, cent2dgy_4)

    L1 = line([cent2dgx_1, cent2dgy_1], [cent2dgx_4, cent2dgy_4])
    L2 = line([cent2dgx_2, cent2dgy_2], [cent2dgx_3, cent2dgy_3])
    R = intersection(L1, L2)
    
    if R:
        shiftx = cx-R[0]
        shifty = cy-R[1]
        if debug: 
            print('\nIntersection coordinates (X,Y):', R[0], R[1], '\n')
            print('Shifts (X,Y):', shiftx, shifty)
        
        if shift:
            array_rec = frame_shift(array, shifty, shiftx, imlib=imlib,
                                    interpolation=interpolation)
            return array_rec, shifty, shiftx
        else:
            return shifty, shiftx
    else:
        print('Something went wrong, no intersection found.')
        return 0


def cube_recenter_satspots(array, xy, subim_size=19, sigfactor=6, debug=False):
    """ Function analog to frame_center_satspots but for image sequences. It 
    actually will call frame_center_satspots for each image in the cube. The
    function also returns the shifted images (not recommended to use when the 
    shifts are of a few percents of a pixel) and plots the histogram of the
    shifts and calculate its statistics. This is important to assess the 
    dispersion of the star center by using artificial waffle/satellite spots
    (like those in VLT/SPHERE images) and evaluate the uncertainty of the 
    position of the center. The use of the shifted images is not recommended.   
    
    Parameters
    ----------
    array : array_like, 3d
        Input cube.
    xy : tuple
        Tuple with coordinates X,Y of the satellite spots in this order:
        upper left, upper right, lower left, lower right. 
    subim_size : int, optional
        Size of subimage where the fitting is done.
    sigfactor : int, optional
        The background pixels will be thresholded before fitting a 2d Gaussian
        to the data using sigma clipped statistics. All values smaller than
        (MEDIAN + sigfactor*STDDEV) will be replaced by small random Gaussian 
        noise. 
    debug : {False, True}, optional 
        If True debug information is printed and plotted (fit and residuals,
        intersections and shifts). This has to be used carefully as it can 
        produce too much output and plots. 
    
    Returns
    ------- 
    array_rec
        The shifted cube.
    shift_y, shift_x
        Shifts Y,X to get to the true center for each image.
    
    """    
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')

    start_time = time_ini()

    n_frames = array.shape[0]
    shift_x = np.zeros((n_frames))
    shift_y = np.zeros((n_frames))
    array_rec = []
    
    bar = pyprind.ProgBar(n_frames, stream=1, title='Looping through frames')
    for i in range(n_frames):
        res = frame_center_satspots(array[i], xy, debug=debug, shift=True,
                                    subim_size=subim_size, sigfactor=sigfactor)
        array_rec.append(res[0])
        shift_y[i] = res[1] 
        shift_x[i] = res[2]          
        bar.update()
       
    timing(start_time)

    plt.figure(figsize=(13,4))
    plt.plot(shift_x, '.-', lw=0.5, color='green', label='Shifts X')
    plt.plot(shift_y, '.-', lw=0.5, color='blue', label='Shifts Y')
    plt.xlim(0, shift_y.shape[0]+5)
    _=plt.xticks(range(0,n_frames,5))
    plt.legend()

    print('AVE X,Y', np.mean(shift_x), np.mean(shift_y))
    print('MED X,Y', np.median(shift_x), np.median(shift_y))
    print('STD X,Y', np.std(shift_x), np.std(shift_y))

    plt.figure()
    b = int(np.sqrt(n_frames))
    _ = plt.hist(shift_x, bins=b, alpha=0.5, color='green', label='Shifts X')
    _ = plt.hist(shift_y, bins=b, alpha=0.5, color='blue', label='Shifts Y')
    plt.legend()

    array_rec = np.array(array_rec) 
    return array_rec, shift_y, shift_x


def frame_center_radon(array, cropsize=101, hsize=0.4, step=0.01,
                       mask_center=None, nproc=None, satspots=False,
                       full_output=False, verbose=True, plot=True, debug=False):
    """ Finding the center of a broadband (co-added) frame with speckles and 
    satellite spots elongated towards the star (center). 
    
    The radon transform comes from scikit-image package. Takes a few seconds to
    compute one radon transform with good resolution. 
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image.
    cropsize : odd int, optional
        Size in pixels of the cropped central area of the input array that will
        be used. It should be large enough to contain the satellite spots.
    hsize : float, optional
        Size of the box for the grid search. The frame is shifted to each 
        direction from the center in a hsize length with a given step.
    step : float, optional
        The step of the coordinates change.
    mask_center : None or int, optional
        If None the central area of the frame is kept. If int a centered zero 
        mask will be applied to the frame. By default the center isn't masked.
    nproc : int, optional
        Number of processes for parallel computing. If None the number of 
        processes will be set to (cpu_count()/2). 
    verbose : {True, False}, bool optional
        Whether to print to stdout some messages and info.
    plot : {True, False}, bool optional
        Whether to plot the radon cost function. 
    debug : {False, True}, bool optional
        Whether to print and plot intermediate info.
    
    Returns
    -------
    optimy, optimx : float
        Values of the Y, X coordinates of the center of the frame based on the
        radon optimization.
    If full_output is True then the radon cost function surface is returned 
    along with the optimal x and y.
        
    Notes
    -----
    The whole idea of this algorithm is based on Pueyo et al. 2014 paper: 
    http://arxiv.org/abs/1409.6388
    
    """
    from .cosmetics import frame_crop
    
    if not array.ndim==2:
        raise TypeError('Input array is not a frame or 2d array')

    if verbose:  start_time = time_ini()
    frame = array.copy()
    frame = frame_crop(frame, cropsize, verbose=False)
    listyx = np.linspace(start=-hsize, stop=hsize, num=2*hsize/step+1, 
                         endpoint=True)
    if not mask_center:
        radint = 0
    else:
        if not isinstance(mask_center, int):
            raise TypeError
        radint = mask_center
    
    coords = [(y,x) for y in listyx for x in listyx]
    cent, _ = frame_center(frame)
           
    frame = get_annulus(frame, radint, cent-radint)
                
    if debug:
        if satspots:
            samples = 10
            theta = np.hstack((np.linspace(start=40, stop=50, num=samples, 
                                           endpoint=False), 
                               np.linspace(start=130, stop=140, num=samples, 
                                           endpoint=False),
                               np.linspace(start=220, stop=230, num=samples, 
                                           endpoint=False),
                               np.linspace(start=310, stop=320, num=samples, 
                                           endpoint=False)))             
            sinogram = radon(frame, theta=theta, circle=True)
            pp_subplots(frame, sinogram)
            print(np.sum(np.abs(sinogram[cent,:])))
        else:
            theta = np.linspace(start=0., stop=360., num=cent*2, endpoint=False)
            sinogram = radon(frame, theta=theta, circle=True)
            pp_subplots(frame, sinogram)
            print(np.sum(np.abs(sinogram[cent,:])))

    if not nproc:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = (cpu_count()/2) 
    pool = Pool(processes=int(nproc))  
    if satspots:
        res = pool.map(EFT, zip(itt.repeat(_radon_costf2), itt.repeat(frame),
                                itt.repeat(cent), itt.repeat(radint), coords))
    else:
        res = pool.map(EFT, zip(itt.repeat(_radon_costf), itt.repeat(frame),
                                itt.repeat(cent), itt.repeat(radint), coords))
    costf = np.array(res)
    pool.close()
        
    if verbose:  
        msg = 'Done {} radon transform calls distributed in {} processes'
        print(msg.format(len(coords), int(nproc)))

    cost_bound = costf.reshape(listyx.shape[0], listyx.shape[0])
    if plot:
        plt.contour(cost_bound, cmap='CMRmap', origin='lower', lw=1, hold='on')
        plt.imshow(cost_bound, cmap='CMRmap', origin='lower', 
                   interpolation='nearest')
        plt.colorbar()
        plt.grid('off')
        plt.show()
        
    #argm = np.argmax(costf) # index of 1st max in 1d cost function 'surface' 
    #optimy, optimx = coords[argm]
    
    # maxima in the 2d cost function surface
    num_max = np.where(cost_bound==cost_bound.max())[0].shape[0]
    ind_maximay, ind_maximax = np.where(cost_bound==cost_bound.max())
    argmy = ind_maximay[int(np.ceil(num_max/2))-1]
    argmx = ind_maximax[int(np.ceil(num_max/2))-1]
    y_grid = np.array(coords)[:,0].reshape(listyx.shape[0], listyx.shape[0])
    x_grid = np.array(coords)[:,1].reshape(listyx.shape[0], listyx.shape[0])
    optimy = y_grid[argmy, 0] 
    optimx = x_grid[0, argmx]  
    
    if verbose: 
        print('Cost function max: {}'.format(costf.max()))
        print('Cost function # maxima: {}'.format(num_max))
        msg = 'Finished grid search radon optimization. Y={:.5f}, X={:.5f}'
        print(msg.format(optimy, optimx))
        timing(start_time)
    
    if full_output:
        return cost_bound, optimy, optimx
    else:
        return optimy, optimx


def _radon_costf(frame, cent, radint, coords):
    """ Radon cost function used in frame_center_radon().
    """
    frame_shifted = frame_shift(frame, coords[0], coords[1])
    frame_shifted_ann = get_annulus(frame_shifted, radint, cent-radint)
    theta = np.linspace(start=0., stop=360., num=frame_shifted_ann.shape[0],    
                    endpoint=False)
    sinogram = radon(frame_shifted_ann, theta=theta, circle=True)
    costf = np.sum(np.abs(sinogram[cent,:]))
    return costf
                
                
def _radon_costf2(frame, cent, radint, coords):
    """ Radon cost function used in frame_center_radon().
    """
    frame_shifted = frame_shift(frame, coords[0], coords[1])
    frame_shifted_ann = get_annulus(frame_shifted, radint, cent-radint)
    samples = 10
    theta = np.hstack((np.linspace(start=40, stop=50, num=samples, endpoint=False), 
                   np.linspace(start=130, stop=140, num=samples, endpoint=False),
                   np.linspace(start=220, stop=230, num=samples, endpoint=False),
                   np.linspace(start=310, stop=320, num=samples, endpoint=False)))
    
    sinogram = radon(frame_shifted_ann, theta=theta, circle=True)
    costf = np.sum(np.abs(sinogram[cent,:]))
    return costf


      
def cube_recenter_radon(array, full_output=False, verbose=True, imlib='opencv',
                        interpolation='lanczos4', **kwargs):
    """ Recenters a cube looping through its frames and calling the 
    ``frame_center_radon`` function.
    
    Parameters
    ----------
    array : array_like
        Input 3d array or cube.
    full_output : {False, True}, bool optional
        If True the recentered cube is returned along with the y and x shifts.
    verbose : {True, False}, bool optional
        Whether to print timing and intermediate information to stdout.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    kwargs : dict
        Optional parameters (keywords and values) can be passed to the
        ``frame_center_radon`` function.
    
    Returns
    -------
    array_rec : array_like
        Recentered cube.
    If full_output is True:
    y, x : 1d array of floats
        Shifts in y and x.
     
    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')

    if verbose:
        start_time = time_ini()

    n_frames = array.shape[0]
    x = np.zeros((n_frames))
    y = np.zeros((n_frames))
    array_rec = array.copy()
    
    bar = pyprind.ProgBar(n_frames, stream=1, title='Looping through frames')
    for i in range(n_frames):
        y[i], x[i] = frame_center_radon(array[i], verbose=False, plot=False, 
                                        **kwargs)
        array_rec[i] = frame_shift(array[i], y[i], x[i], imlib=imlib,
                                   interpolation=interpolation)
        bar.update()
        
    if verbose:
        timing(start_time)

    if full_output:
        return array_rec, y, x
    else:
        return array_rec


# TODO: verify correct handling of even/odd cases
def cube_recenter_dft_upsampling(array, cy_1, cx_1, negative=False, fwhm=4, 
                                 subi_size=None, upsample_factor=100,
                                 imlib='opencv', interpolation='lanczos4',
                                 full_output=False, verbose=True,
                                 save_shifts=False, debug=False):
    """ Recenters a cube of frames using the DFT upsampling method as 
    proposed in Guizar et al. 2008 (see Notes) plus a chi^2, for determining
    automatically the upsampling factor, as implemented in the package 
    'image_registration' (see Notes).
    
    The algorithm (DFT upsampling) obtains an initial estimate of the 
    cross-correlation peak by an FFT and then refines the shift estimation by 
    upsampling the DFT only in a small neighborhood of that estimate by means 
    of a matrix-multiply DFT.
    
    Parameters
    ----------
    array : array_like
        Input cube.
    cy_1, cx_1 : int
        Coordinates of the center of the subimage for fitting a 2d Gaussian and
        centroiding the 1st frame. 
    negative : {False, True}, optional
        If True the centroiding of the 1st frames is done with a negative 
        2d Gaussian fit.   
    fwhm : float, optional
        FWHM size in pixels.
    subi_size : int or None, optional
        Size of the square subimage sides in terms of FWHM that will be used
        to centroid to frist frame. If subi_size is None then the first frame
        is assumed to be centered already.
    upsample_factor :  int optional
        Upsampling factor (default 100). Images will be registered to within
        1/upsample_factor of a pixel.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    full_output : {False, True}, bool optional
        Whether to return 2 1d arrays of shifts along with the recentered cube 
        or not.
    verbose : {True, False}, bool optional
        Whether to print to stdout the timing or not.
    save_shifts : {False, True}, bool optional
        Whether to save the shifts to a file in disk.
    debug : {False, True}, bool optional
        Whether to print to stdout the shifts or not. 
    
    Returns
    -------
    array_recentered : array_like
        The recentered cube. Frames have now odd size.
    If full_output is True:
    y, x : array_like
        1d arrays with the shifts in y and x.     
    
    Notes
    -----
    Using the implementation from skimage.feature.register_translation.
    
    Guizar-Sicairos et al. "Efficient subpixel image registration algorithms," 
    Opt. Lett. 33, 156-158 (2008). 
    The algorithm registers two images (2-D rigid translation) within a fraction 
    of a pixel specified by the user. Instead of computing a zero-padded FFT
    (fast Fourier transform), this code uses selective upsampling by a
    matrix-multiply DFT (discrete FT) to dramatically reduce computation time
    and memory without sacrificing accuracy. With this procedure all the image
    points are used to compute the upsampled cross-correlation in a very small
    neighborhood around its peak.
    
    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')

    if verbose:
        start_time = time_ini()
    
    n_frames = array.shape[0]
    x = np.zeros((n_frames))
    y = np.zeros((n_frames))
    array_rec = array.copy()

    cy, cx = frame_center(array[0])
    # Centroiding first frame with 2d gaussian and shifting
    if subi_size is not None:
        size = int(np.round(fwhm*subi_size))
        y1, x1 = _centroid_2dg_frame(array_rec, 0, size, cy_1, cx_1, negative,
                                     debug=debug)
        x[0] = cx-x1
        y[0] = cy-y1
        array_rec[0] = frame_shift(array_rec[0], shift_y=y[0], shift_x=x[0],
                                   imlib=imlib, interpolation=interpolation)
        if verbose:
            print("\nShift for first frame X,Y=({:.3f},{:.3f})".format(x[0],y[0]))
            print("The rest of the frames will be shifted by cross-correlation" \
                  " with the first one")
        if debug:
            pp_subplots(frame_crop(array[0], size, verbose=False),
                        frame_crop(array_rec[0], size, verbose=False),
                        grid=True, title='original / shifted 1st frame subimage')
    else:
        if verbose:
            print("It's assumed that the first frame is well centered")
            print("The rest of the frames will be shifted by cross-correlation" \
                  " with the first one")
        x[0] = cx
        y[0] = cy
    
    # Finding the shifts with DTF upsampling of each frame wrt the first
    bar = pyprind.ProgBar(n_frames, stream=1, title='Looping through frames')
    for i in range(1, n_frames):
        shift_yx, _, _ = register_translation(array_rec[0], array[i],
                                              upsample_factor=upsample_factor)
        y[i], x[i] = shift_yx
        array_rec[i] = frame_shift(array[i], shift_y=y[i], shift_x=x[i],
                                   imlib=imlib, interpolation=interpolation)
        bar.update()

    if debug:
        print("\nShifts in X and Y")
        for i in range(n_frames):
            print(x[i], y[i])
        
    if verbose:
        timing(start_time)
        
    if save_shifts: 
        np.savetxt('recent_dft_shifts.txt', np.transpose([y, x]), fmt='%f')
    if full_output:
        return array_rec, y, x
    else:
        return array_rec
  

def cube_recenter_2dfit(array, xy=None, fwhm=4, subi_size=5, model='gauss',
                        nproc=1, imlib='opencv', interpolation='lanczos4',
                        offset=None, negative=False, threshold=False,
                        save_shifts=False, full_output=False, verbose=True,
                        debug=False,):
    """ Recenters the frames of a cube. The shifts are found by fitting a 2d 
    Gaussian or Moffat to a subimage centered at ``xy``. This assumes the frames
    don't have too large shifts (>5px). The frames are shifted using the 
    function frame_shift().
    
    Parameters
    ----------
    array : array_like
        Input cube.
    xy : tuple of int
        Coordinates of the center of the subimage (wrt the original frame).
    fwhm : float or array_like
        FWHM size in pixels, either one value (float) that will be the same for
        the whole cube, or an array of floats with the same dimension as the 
        0th dim of array, containing the fwhm for each channel (e.g. in the case
        of an ifs cube, where the fwhm varies with wavelength)
    subi_size : int, optional
        Size of the square subimage sides in pixels.
    mode : str, optional
        Sets the type of fit to be used. 'gauss' for a 2d Gaussian fit and
        'moff' for a 2d Moffat fit.
    nproc : int or None, optional
        Number of processes (>1) for parallel computing. If 1 then it runs in 
        serial. If None the number of processes will be set to (cpu_count()/2).
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    offset : tuple of floats, optional
        If None the region of the frames used for the 2d Gaussian fit is shifted
        to the center of the images (2d arrays). If a tuple is given it serves
        as the offset of the fitted area wrt the center of the 2d arrays.
    negative : bool, optional
        If True a negative 2d Gaussian/Moffat fit is performed.
    threshold : bool, optional
        If True the background pixels (estimated using sigma clipped statistics)
        will be replaced by small random Gaussian noise.
    save_shifts : bool, optional
        Whether to save the shifts to a file in disk.
    full_output : bool, optional
        Whether to return 2 1d arrays of shifts along with the recentered cube 
        or not.
    verbose : bool, optional
        Whether to print to stdout the timing or not.
    debug : bool, optional
        If True the details of the fitting are shown. Won't work when the cube
        contains >20 frames (as it might produce an extremely long output).
        
    Returns
    -------
    array_recentered : array_like
        The recentered cube. Frames have now odd size.
    If full_output is True:
    y, x : array_like
        1d arrays with the shifts in y and x. 
    
    """
    if verbose:
        start_time = time_ini()

    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')

    n_frames, y, x = array.shape

    if not isinstance(subi_size, int):
        raise ValueError('subi_size must be an integer')
    if y % 2 == 0:
        if subi_size % 2 != 0:
            subi_size += 1
            print('subi_size is odd (while frame size is even)')
            print('Setting subi_size to {} pixels'.format(subi_size))
    else:
        if subi_size % 2 == 0:
            subi_size += 1
            print('subi_size is even (while frame size is odd)')
            print('Setting subi_size to {} pixels'.format(subi_size))

    if isinstance(fwhm, (float, int)):
        fwhm = np.ones(n_frames) * fwhm

    if debug and array.shape[0] > 20:
        msg = 'Debug with a big array will produce a very long output. '
        msg += 'Try with less than 20 frames in debug mode'
        raise RuntimeWarning(msg)

    if xy is not None:
        pos_x, pos_y = xy
        if not isinstance(pos_x, int) or not isinstance(pos_y, int):
            raise TypeError('pos_x and pos_y must be integers')
    else:
        pos_y, pos_x = frame_center(array[0])

    cy, cx = frame_center(array[0])
    array_recentered = np.empty_like(array)

    if model == 'gauss':
        func = _centroid_2dg_frame
    elif model == 'moff':
        func = _centroid_2dm_frame
    else:
        raise ValueError('model not recognized')
    
    if not nproc:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = (cpu_count()/2) 
    if nproc == 1:
        res = []
        bar = pyprind.ProgBar(n_frames, stream=1, 
                              title='2d Gauss-fitting, looping through frames')
        for i in range(n_frames):
            res.append(func(array, i, subi_size, pos_y, pos_x, negative, debug,
                            fwhm[i], threshold))
            bar.update()
        res = np.array(res)
    elif nproc > 1:
        pool = Pool(processes=int(nproc))  
        res = pool.map(EFT, zip(itt.repeat(func), itt.repeat(array),
                                range(n_frames), itt.repeat(subi_size),
                                itt.repeat(pos_y), itt.repeat(pos_x),
                                itt.repeat(negative), itt.repeat(debug), fwhm,
                                itt.repeat(threshold)))
        res = np.array(res)
        pool.close()
    y = cy - res[:, 0]
    x = cx - res[:, 1]

    if offset is not None:
        offx, offy = offset
        y -= offy
        x -= offx
    
    bar2 = pyprind.ProgBar(n_frames, stream=1, title='Shifting the frames')
    for i in range(n_frames):
        if debug:
            print("\nShifts in X and Y")
            print(x[i], y[i])
        array_recentered[i] = frame_shift(array[i], y[i], x[i], imlib=imlib,
                                          interpolation=interpolation)
        bar2.update()
        
    if verbose:
        timing(start_time)

    if save_shifts: 
        np.savetxt('recent_gauss_shifts.txt', np.transpose([y, x]), fmt='%f')
    if full_output:
        return array_recentered, y, x
    else:
        return array_recentered


def _centroid_2dg_frame(cube, frnum, size, pos_y, pos_x, negative, debug,
                        fwhm, threshold=False):
    """ Finds the centroid by using a 2d gaussian fitting in one frame from a
    cube.
    """
    sub_image, y1, x1 = get_square(cube[frnum], size=size, y=pos_y, x=pos_x,
                                   position=True)
    # negative gaussian fit
    if negative:
        sub_image = -sub_image + np.abs(np.min(-sub_image))

    y_i, x_i = fit_2dgaussian(sub_image, crop=False, fwhmx=fwhm, fwhmy=fwhm,
                              threshold=threshold, sigfactor=1, debug=debug)
    y_i = y1 + y_i
    x_i = x1 + x_i
    return y_i, x_i


def _centroid_2dm_frame(cube, frnum, size, pos_y, pos_x, negative, debug,
                        fwhm, threshold=False):
    """ Finds the centroid by using a 2d moffat fitting in one frame from a
    cube.
    """
    sub_image, y1, x1 = get_square(cube[frnum], size=size, y=pos_y, x=pos_x,
                                   position=True)
    sub_image = sub_image.byteswap().newbyteorder()
    # negative fit
    if negative:
        sub_image = -sub_image + np.abs(np.min(-sub_image))

    y_i, x_i = fit_2dmoffat(sub_image, y1, x1, full_output=False)
    return y_i, x_i


# TODO: verify correct handling of even/odd cases
def cube_recenter_via_speckles(cube_sci, cube_ref=None, alignment_iter=5,
                               gammaval=1, min_spat_freq=0.5, max_spat_freq=3,
                               fwhm=4, debug=False, negative=True,
                               recenter_median=True, subframesize=151,
                               imlib='opencv', interpolation='bilinear'):
    """ Registers frames based on the median speckle pattern. Optionally centers 
    based on the position of the vortex null in the median frame. Images are 
    filtered to isolate speckle spatial frequencies.

    Parameters
    ----------
    cube_sci : array_like  
        Science cube.
    cube_ref : array_like
        Reference cube (e.g. for NIRC2 data in RDI mode). 
    alignment_iter : int, optional  
        Number of alignment iterations (recomputes median after each iteration). 
    gammaval : int, optional 
        Applies a gamma correction to emphasize speckles (useful for faint 
        stars).
    min_spat_freq : float, optional 
        Spatial frequency for high pass filter. 
    max_spat_freq : float, optional 
        Spatial frequency for low pass filter. 
    fwhm : float, optional 
        Full width at half maximum. 
    debug : {False, True}, optional 
        Outputs extra info.
    negative : {True, False}, optional
        Use a negative gaussian fit to determine the center of the median frame.
    recenter_median : {True, False}, optional 
        Recenter the frames at each iteration based on the gaussian fit.
    subframesize : int, optional
        Sub-frame window size used. Should cover the region where speckles are 
        the dominant noise source. 
    imlib : str, optional 
        Image processing library to use. 
    interpolation : str, optional 
        Interpolation method to use.


    Returns
    -------
    array_shifted : array_like
        Shifted 2d array.

    If cube_ref is not None, returns:
    cube_reg_sci: Registered science cube.
    cube_reg_ref: Ref. cube registered to science frames.
    cum_x_shifts_sci: Vector of x shifts for science frames.
    cum_y_shifts_sci: Vector of y shifts for science frames.
    cum_x_shifts_ref: Vector of x shifts for ref. frames.
    cum_y_shifts_ref: Vector of y shifts for ref. frames.

    Otherwise, returns: cube_reg_sci, cum_x_shifts_sci, cum_y_shifts_sci
    """
    if cube_ref is not None:
        refStar = True
    else:
        refStar = False

    cube_sci_subframe = cube_crop_frames(cube_sci, subframesize, verbose=False)
    if (refStar): cube_ref_subframe = cube_crop_frames(cube_ref, subframesize,
                                                       verbose=False)

    ceny, cenx = frame_center(cube_sci_subframe[0, :, :])
    print('sub frame is ' + str(cube_sci_subframe.shape[1]) + 'x' + str(
        cube_sci_subframe.shape[2]))
    print('center pixel is (' + str(ceny) + ', ' + str(cenx) + ')')

    # Make a copy of the sci and ref frames, filter them.
    # will be used for alignment purposes
    cube_sci_lpf = cube_sci_subframe.copy()
    if refStar:
        cube_ref_lpf = cube_ref_subframe.copy()

    cube_sci_lpf = cube_sci_lpf - np.min(cube_sci_lpf)
    if refStar:
        cube_ref_lpf = cube_ref_lpf - np.min(cube_ref_lpf)

    # Remove spatial frequencies <0.5 lam/D and >3lam/D to isolate speckles
    for i in range(cube_sci.shape[0]):
        cube_sci_lpf[i, :, :] = cube_sci_lpf[i, :, :] - frame_filter_lowpass(
            cube_sci_lpf[i, :, :], 'median', median_size=fwhm * max_spat_freq)
    if refStar:
        for i in range(cube_ref.shape[0]):
            cube_ref_lpf[i, :, :] = cube_ref_lpf[i, :,
                                    :] - frame_filter_lowpass(
                cube_ref_lpf[i, :, :], 'median',
                median_size=fwhm * max_spat_freq)

    for i in range(cube_sci.shape[0]):
        cube_sci_lpf[i, :, :] = frame_filter_gaussian2d(cube_sci_lpf[i, :, :],
                                                        min_spat_freq * fwhm)
    if refStar:
        for i in range(cube_ref.shape[0]):
            cube_ref_lpf[i, :, :] = frame_filter_gaussian2d(
                cube_ref_lpf[i, :, :], min_spat_freq * fwhm)

    if refStar:
        alignment_cube = np.zeros((1 + cube_sci.shape[0] + cube_ref.shape[0],
                                   cube_sci_subframe.shape[1],
                                   cube_sci_subframe.shape[2]))
        alignment_cube[1:(cube_sci.shape[0] + 1), :, :] = cube_sci_lpf
        alignment_cube[
        (cube_sci.shape[0] + 1):(cube_sci.shape[0] + 2 + cube_ref.shape[0]), :,
        :] = cube_ref_lpf
    else:
        alignment_cube = np.zeros((1 + cube_sci.shape[0],
                                   cube_sci_subframe.shape[1],
                                   cube_sci_subframe.shape[2]))
        alignment_cube[1:(cube_sci.shape[0] + 1), :, :] = cube_sci_lpf

    n_frames = alignment_cube.shape[0]  # number of sci+ref frames + 1 for the median

    cum_y_shifts = 0
    cum_x_shifts = 0

    for i in range(alignment_iter):
        alignment_cube[0] = np.median(alignment_cube[1:(cube_sci.shape[0] + 1), :, :], axis=0)
        if recenter_median:
            # Recenter the median frame using a neg. gaussian fit
            sub_image, y1, x1 = get_square(alignment_cube[0], size=int(fwhm),
                                           y=ceny, x=cenx, position=True)
            if negative:
                sub_image = -sub_image + np.abs(np.min(-sub_image))
            y_i, x_i = fit_2dgaussian(sub_image, crop=False, threshold=False,
                                      sigfactor=1, debug=debug)
            yshift = ceny - (y1 + y_i)
            xshift = cenx - (x1 + x_i)

            alignment_cube[0, :, :] = frame_shift(alignment_cube[0, :, :],
                                                  yshift, xshift, imlib=imlib,
                                                  interpolation=interpolation)

        # center the cube with stretched values
        cube_stret = np.log10((abs(alignment_cube) + 1) ** (gammaval))
        _, y_shift, x_shift = cube_recenter_dft_upsampling(cube_stret, ceny,
                                                           cenx, fwhm=fwhm,
                                                           subi_size=None,
                                                           full_output=True,
                                                           verbose=False)

        print('\nSquare sum of shift vecs: ' + str(
            np.sum(np.sqrt(y_shift ** 2 + x_shift ** 2))))

        for i in range(1, n_frames):
            alignment_cube[i] = frame_shift(alignment_cube[i], y_shift[i],
                                            x_shift[i], imlib=imlib,
                                            interpolation=interpolation)

        cum_y_shifts = cum_y_shifts + y_shift
        cum_x_shifts = cum_x_shifts + x_shift

    cum_y_shifts_sci = cum_y_shifts[1:(cube_sci.shape[0] + 1)]
    cum_x_shifts_sci = cum_x_shifts[1:(cube_sci.shape[0] + 1)]

    cube_reg_sci = cube_sci.copy()
    for i in range(cube_sci.shape[0]):
        cube_reg_sci[i] = frame_shift(cube_sci[i], cum_y_shifts_sci[i],
                                      cum_x_shifts_sci[i], imlib=imlib,
                                      interpolation=interpolation)

    if refStar:
        cube_reg_ref = cube_ref.copy()
        cum_y_shifts_ref = cum_y_shifts[(cube_sci.shape[0] + 1):]
        cum_x_shifts_ref = cum_x_shifts[(cube_sci.shape[0] + 1):]
        for i in range(cube_ref.shape[0]):
            cube_reg_ref[i] = frame_shift(cube_ref[i], cum_y_shifts_ref[i],
                                          cum_x_shifts_ref[i], imlib=imlib,
                                          interpolation=interpolation)

    if cube_ref is not None:
        return cube_reg_sci, cube_reg_ref, cum_x_shifts_sci, cum_y_shifts_sci, cum_x_shifts_ref, cum_y_shifts_ref
    else:
        return cube_reg_sci, cum_x_shifts_sci, cum_y_shifts_sci




        



#! /usr/bin/env python

"""
Module with the MCMC (``emcee``) sampling for NEGFC parameter estimation.
"""
from __future__ import print_function

__author__ = 'O. Wertz, Carlos Alberto Gomez Gonzalez'
__all__ = ['mcmc_negfc_sampling',
           'chain_zero_truncated',
           'show_corner_plot',
           'show_walk_plot',
           'confidence']

import numpy as np
import os
import emcee
from math import isinf, floor, ceil
import inspect
import datetime
import corner
from matplotlib import pyplot as plt
from matplotlib.ticker import MaxNLocator
from matplotlib.mlab import normpdf
from scipy.stats import norm
from ..fits import open_adicube, open_fits
from ..phot import cube_inject_companions
from ..conf import time_ini, timing
from ..conf.utils_conf import sep
from .simplex_fmerit import get_values_optimize
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)



def lnprior(param, bounds):
    """ Define the prior log-function.
    
    Parameters
    ----------    
    param: tuple
        The model parameters.
    bounds: list
        The bounds for each model parameter. 
        Ex: bounds = [(10,20),(0,360),(0,5000)]        
    
    Returns
    -------
    out: float. 
        0 if all the model parameters satisfy the prior conditions defined here.
        -np.inf if at least one model parameters is out of bounds.
    """
    
    try:
        r, theta, flux = param
    except TypeError:
        print('paraVector must be a tuple, {} was given'.format(type(param)))

    try:
        r_bounds, theta_bounds, flux_bounds = bounds
    except TypeError:
        print('bounds must be a list of tuple, {} was given'.format(type(param)))
        
    if r_bounds[0] <= r <= r_bounds[1] and \
       theta_bounds[0] <= theta <= theta_bounds[1] and \
       flux_bounds[0] <= flux <= flux_bounds[1]:
        return 0.0
    else:
        return -np.inf


def lnlike(param, cube, angs, plsc, psf_norm, fwhm, annulus_width,
           ncomp, aperture_radius, initial_state, cube_ref=None,
           svd_mode='lapack', scaling='temp-mean', fmerit='sum', imlib='opencv',
           interpolation='lanczos4', collapse='median', debug=False):
    """ Define the likelihood log-function.
    
    Parameters
    ----------    
    param: tuple
        The model parameters, typically (r, theta, flux).
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array. 
    plsc: float
        The platescale, in arcsec per pixel.
    psf_norm: numpy.array
        The scaled psf expressed as a numpy.array.    
    annulus_width: float
        The width of the annulus of interest in terms of the FWHM.
    ncomp: int
        The number of principal components.
    fwhm : float
        The FHWM in pixels.
    aperture_radius: float
        The radius of the circular aperture in terms of the FWHM. 
    initial_state: numpy.array
        The initial guess for the position and the flux of the planet.
    cube_ref: array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Chooses the figure of merit to be used. stddev works better for close in
        companions sitting on top of speckle noise.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    debug: boolean
        If True, the cube is returned along with the likelihood log-function.        
        
    Returns
    -------
    out: float
        The log of the likelihood.
        
    """    
    # Create the cube with the negative fake companion injected
    cube_negfc = cube_inject_companions(cube, psf_norm, angs, flevel=-param[2],
                                        plsc=plsc, rad_dists=[param[0]],
                                        n_branches=1, theta=param[1],
                                        imlib=imlib, interpolation=interpolation,
                                        verbose=False)
                                  
    # Perform PCA and extract the zone of interest
    values = get_values_optimize(cube_negfc,angs,ncomp,annulus_width*fwhm,
                                 aperture_radius*fwhm, initial_state[0],
                                 initial_state[1], cube_ref=cube_ref,
                                 svd_mode=svd_mode, scaling=scaling,
                                 imlib=imlib, interpolation=interpolation,
                                 collapse=collapse)
    
    # Function of merit
    if fmerit == 'sum':
        lnlikelihood = -0.5 * np.sum(np.abs(values))
    elif fmerit == 'stddev':
        values = values[values != 0]
        lnlikelihood = -1*np.std(np.abs(values))
    else:
        raise RuntimeError('fmerit choice not recognized.')
    
    if debug:
        return lnlikelihood, cube_negfc
    else:
        return lnlikelihood


def lnprob(param,bounds, cube, angs, plsc, psf_norm, fwhm,
           annulus_width, ncomp, aperture_radius, initial_state, cube_ref=None,
           svd_mode='lapack', scaling='temp-mean', fmerit='sum', imlib='opencv',
           interpolation='lanczos4', collapse='median', display=False):
    """ Define the probability log-function as the sum between the prior and 
    likelihood log-funtions.
    
    Parameters
    ----------    
    param: tuple
        The model parameters.
    bounds: list
        The bounds for each model parameter. 
        Ex: bounds = [(10,20),(0,360),(0,5000)] 
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array. 
    plsc: float
        The platescale, in arcsec per pixel.
    psf_norm: numpy.array
        The scaled psf expressed as a numpy.array.   
    fwhm : float
        The FHWM in pixels. 
    annulus_width: float
        The width in pixel of the annulus on wich the PCA is performed.
    ncomp: int
        The number of principal components.
    aperture_radius: float
        The radius of the circular aperture.  
    initial_state: numpy.array
        The initial guess for the position and the flux of the planet. 
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Chooses the figure of merit to be used. stddev works better for close in
        companions sitting on top of speckle noise.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    display: boolean
        If True, the cube is displayed with ds9.        
        
    Returns
    -------
    out: float
        The probability log-function.
    
    """      
    if initial_state is None:
        initial_state = param
    
    lp = lnprior(param, bounds)
    
    if isinf(lp):
        return -np.inf       
    
    return lp + lnlike(param, cube, angs, plsc, psf_norm, fwhm,
                       annulus_width, ncomp, aperture_radius, initial_state,
                       cube_ref, svd_mode, scaling, fmerit, imlib,
                       interpolation, collapse, display)


def gelman_rubin(x):      
    """
    Determine the Gelman-Rubin \hat{R} statistical test between Markov chains.
    
    Parameters
    ----------
    x: numpy.array
        The numpy.array on which the Gelman-Rubin test is applied. This array
        should contain at least 2 set of data, i.e. x.shape >= (2,).
        
    Returns
    -------
    out: float
        The Gelman-Rubin \hat{R}.

    Example
    -------
    >>> x1 = np.random.normal(0.0,1.0,(1,100))
    >>> x2 = np.random.normal(0.1,1.3,(1,100))
    >>> x = np.vstack((x1,x2))
    >>> gelman_rubin(x)
    1.0366629898991262
    >>> gelman_rubin(np.vstack((x1,x1)))
    0.99
        
    """
    if np.shape(x) < (2,):
        raise ValueError(
            'Gelman-Rubin diagnostic requires multiple chains of the same length.')

    try:
        m, n = np.shape(x)
    except ValueError:
        print("Bad shape for the chains")
        return

    # Calculate between-chain variance
    B_over_n = np.sum((np.mean(x, 1) - np.mean(x)) ** 2) / (m - 1)

    # Calculate within-chain variances
    W = np.sum([(x[i] - xbar) ** 2 for i, xbar in enumerate(np.mean(x,
                                                           1))]) / (m * (n - 1))
    # (over) estimate of variance
    s2 = W * (n - 1) / n + B_over_n

    # Pooled posterior variance estimate
    V = s2 + B_over_n / m

    # Calculate PSRF
    R = V / W

    return R


def gelman_rubin_from_chain(chain, burnin):
    """
    Pack the MCMC chain and determine the Gelman-Rubin \hat{R} statistical test.
    In other words, two sub-sets are extracted from the chain (burnin parts are
    taken into account) and the Gelman-Rubin statistical test is performed.
    
    Parameters
    ----------
    chain: numpy.array
        The MCMC chain with the shape walkers x steps x model_parameters
    burnin: float \in [0,1]
        The fraction of a walker which is discarded.
        
    Returns
    -------
    out: float
        The Gelman-Rubin \hat{R}.
        
    """
    dim = chain.shape[2]
    k = chain.shape[1]
    
    threshold0 = int(floor(burnin*k))
    threshold1 = int(floor((1-burnin)*k*0.25))   
    
    rhat = np.zeros(dim)
    
    for j in range(dim):
        part1 = chain[:,threshold0:threshold0+threshold1,j].reshape((-1))
        part2 = chain[:,threshold0+3*threshold1:threshold0+4*threshold1,j].reshape((-1))
        series = np.vstack((part1,part2))
        rhat[j] = gelman_rubin(series)
        
    return rhat


def mcmc_negfc_sampling(cubes, angs, psfn, ncomp, plsc, initial_state, fwhm=4,
                        annulus_width=3, aperture_radius=4, cube_ref=None,
                        svd_mode='lapack', scaling='temp-mean', fmerit='sum',
                        imlib='opencv', interpolation='lanczos4',
                        collapse='median', nwalkers=1000, bounds=None, a=2.0,
                        burnin=0.3, rhat_threshold=1.01, rhat_count_threshold=1,
                        niteration_min=0, niteration_limit=1e02, 
                        niteration_supp=0, check_maxgap=1e04, nproc=1, 
                        output_file=None, display=False, verbose=True,
                        save=False):
    """ Runs an affine invariant mcmc sampling algorithm in order to determine
    the position and the flux of the planet using the 'Negative Fake Companion'
    technique. The result of this procedure is a chain with the samples from the
    posterior distributions of each of the 3 parameters.
    
    This technique can be summarized as follows:
    
    1)  We inject a negative fake companion (one candidate) at a given 
        position and characterized by a given flux, both close to the expected 
        values.
    2)  We run PCA on an full annulus which pass through the initial guess, 
        regardless of the position of the candidate.
    3)  We extract the intensity values of all the pixels contained in a 
        circular aperture centered on the initial guess.
    4)  We calculate the function of merit. The associated chi^2 is given by
        chi^2 = sum(|I_j|) where j \in {1,...,N} with N the total number of 
        pixels contained in the circular aperture.        
    The steps 1) to 4) are looped. At each iteration, the candidate model 
    parameters are defined by the emcee Affine Invariant algorithm. 
    
    Parameters
    ----------  
    cubes: str or numpy.array
        The relative path to the cube of fits images OR the cube itself.
    angs: str or numpy.array
        The relative path to the parallactic angle fits image or the angs itself.
    psfn: str or numpy.array
        The relative path to the instrumental PSF fits image or the PSF itself.
        The PSF must be centered and the flux in a 1*FWHM aperture must equal 1.
    ncomp: int
        The number of principal components.        
    plsc: float
        The platescale, in arcsec per pixel.  
    annulus_width: float, optional
        The width in pixel of the annulus on which the PCA is performed.
    aperture_radius: float, optional
        The radius of the circular aperture.        
    nwalkers: int optional
        The number of Goodman & Weare 'walkers'.
    initial_state: numpy.array
        The first guess for the position and flux of the planet, respectively.
        Each walker will start in a small ball around this preferred position.
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
        'randsvd' is not recommended for the negative fake companion technique.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Chooses the figure of merit to be used. stddev works better for close in
        companions sitting on top of speckle noise.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    bounds: numpy.array or list, default=None, optional
        The prior knowledge on the model parameters. If None, large bounds will 
        be automatically estimated from the initial state.
    a: float, default=2.0
        The proposal scale parameter. See notes.
    burnin: float, default=0.3
        The fraction of a walker which is discarded.
    rhat_threshold: float, default=0.01
        The Gelman-Rubin threshold used for the test for nonconvergence.   
    rhat_count_threshold: int, optional
        The Gelman-Rubin test must be satisfied 'rhat_count_threshold' times in
        a row before claiming that the chain has converged.        
    niteration_min: int, optional
        Steps per walker lower bound. The simulation will run at least this
        number of steps per walker.
    niteration_limit: int, optional
        Steps per walker upper bound. If the simulation runs up to 
        'niteration_limit' steps without having reached the convergence 
        criterion, the run is stopped.
    niteration_supp: int, optional
        Number of iterations to run after having "reached the convergence".     
    check_maxgap: int, optional
        Maximum number of steps per walker between two Gelman-Rubin test.
    nproc: int, optional
        The number of processes to use for parallelization. 
    output_file: str
        The name of the ouput file which contains the MCMC results 
        (if save is True).
    display: boolean
        If True, the walk plot is displayed at each evaluation of the Gelman-
        Rubin test.
    verbose: boolean
        Display informations in the shell.
    save: boolean
        If True, the MCMC results are pickled.
                    
    Returns
    -------
    out : numpy.array
        The MCMC chain.         
        
    Notes
    -----
    The parameter 'a' must be > 1. For more theoretical information concerning
    this parameter, see Goodman & Weare, 2010, Comm. App. Math. Comp. Sci., 
    5, 65, Eq. [9] p70.
    
    The parameter 'rhat_threshold' can be a numpy.array with individual 
    threshold value for each model parameter.
    """ 
    if verbose:
        start_time = time_ini()
        print("        MCMC sampler for the NEGFC technique       ")
        print(sep)

    # If required, one create the output folder.    
    if save:    
        if not os.path.exists('results'):
            os.makedirs('results')
        
        if output_file is None:
            datetime_today = datetime.datetime.today()
            output_file = str(datetime_today.year)+str(datetime_today.month)+\
                          str(datetime_today.day)+'_'+str(datetime_today.hour)+\
                          str(datetime_today.minute)+str(datetime_today.second)            
        
        if not os.path.exists('results/'+output_file):
            os.makedirs('results/'+output_file)

            
    # #########################################################################
    # If required, one opens the source files
    # #########################################################################
    if isinstance(cubes,str) and isinstance(angs,str):
        if angs is None:
            cubes, angs = open_adicube(cubes, verbose=False)
        else:
            cubes = open_fits(cubes)
            angs = open_fits(angs, verbose=False)    
        
        if isinstance(psfn,str):
            psfn = open_fits(psfn)
        
        if verbose:
            print('The data has been loaded. Let''s continue !')
    
    # #########################################################################
    # Initialization of the variables
    # #########################################################################    
    dim = 3 # There are 3 model parameters, resp. the radial and angular 
            # position of the planet and its flux.
    
    itermin = niteration_min
    limit = niteration_limit    
    supp = niteration_supp
    maxgap = check_maxgap
    initial_state = np.array(initial_state)
    
    if itermin > limit:
        itermin = 0
        print("'niteration_min' must be < 'niteration_limit'.")
        
    fraction = 0.3
    geom = 0
    lastcheck = 0
    konvergence = np.inf
    rhat_count = 0
        
    chain = np.empty([nwalkers,1,dim])
    isamples = np.empty(0)
    pos = initial_state + np.random.normal(0,1e-01,(nwalkers,3))
    nIterations = limit + supp
    rhat = np.zeros(dim)  
    stop = np.inf
    

    if bounds is None:
        bounds = [(initial_state[0]-annulus_width/2.,initial_state[0]+annulus_width/2.), #radius
                  (initial_state[1]-10,initial_state[1]+10), #angle
                  (0,2*initial_state[2])] #flux
    
    sampler = emcee.EnsembleSampler(nwalkers, dim, lnprob, a,
                                    args =([bounds, cubes, angs, plsc, psfn,
                                            fwhm, annulus_width, ncomp,
                                            aperture_radius, initial_state,
                                            cube_ref, svd_mode, scaling, fmerit,
                                            imlib, interpolation, collapse]),
                                    threads=nproc)
                                    
    start = datetime.datetime.now()

    # #########################################################################
    # Affine Invariant MCMC run
    # ######################################################################### 
    if verbose:
        print('')
        print('Start of the MCMC run ...')
        print('Step  |  Duration/step (sec)  |  Remaining Estimated Time (sec)')
                             
    for k, res in enumerate(sampler.sample(pos,iterations=nIterations,
                                           storechain=True)):
        elapsed = (datetime.datetime.now()-start).total_seconds()
        if verbose:
            if k == 0:
                q = 0.5
            else:
                q = 1
            print('{}\t\t{:.5f}\t\t\t{:.5f}'.format(k,elapsed*q,elapsed*(limit-k-1)*q))
            
        start = datetime.datetime.now()

        # ---------------------------------------------------------------------        
        # Store the state manually in order to handle with dynamical sized chain.
        # ---------------------------------------------------------------------    
        ## Check if the size of the chain is long enough.
        s = chain.shape[1]
        if k+1 > s: #if not, one doubles the chain length
            empty = np.zeros([nwalkers,2*s,dim])
            chain = np.concatenate((chain,empty),axis=1)
        ## Store the state of the chain
        chain[:,k] = res[0]
        
        
        # ---------------------------------------------------------------------
        # If k meets the criterion, one tests the non-convergence.
        # ---------------------------------------------------------------------              
        criterion = np.amin([ceil(itermin*(1+fraction)**geom),\
                            lastcheck+floor(maxgap)])
   
        if k == criterion:
            if verbose:
                print('')
                print('   Gelman-Rubin statistic test in progress ...') 
            
            geom += 1
            lastcheck = k
            if display:
                show_walk_plot(chain)
                
            if save:
                import pickle                                    
                
                with open('results/'+output_file+'/'+output_file+'_temp_k{}'.format(k),'wb') as fileSave:
                    myPickler = pickle.Pickler(fileSave)
                    myPickler.dump({'chain':sampler.chain, 
                                    'lnprob':sampler.lnprobability, 
                                    'AR':sampler.acceptance_fraction})
                
            ## We only test the rhat if we have reached the minimum number of steps.
            if (k+1) >= itermin and konvergence == np.inf:
                threshold0 = int(floor(burnin*k))
                threshold1 = int(floor((1-burnin)*k*0.25))

                # We calculate the rhat for each model parameter.
                for j in range(dim):
                    part1 = chain[:,threshold0:threshold0+threshold1,j].reshape((-1))
                    part2 = chain[:,threshold0+3*threshold1:threshold0+4*threshold1,j].reshape((-1))
                    series = np.vstack((part1,part2))
                    rhat[j] = gelman_rubin(series)   
                if verbose:    
                    print('   r_hat = {}'.format(rhat))
                    print('   r_hat <= threshold = {}'.format(rhat <= rhat_threshold))
                    print('')
                # We test the rhat.
                if (rhat <= rhat_threshold).all(): #and rhat_count < rhat_count_threshold: 
                    rhat_count += 1
                    if rhat_count < rhat_count_threshold:
                        print("Gelman-Rubin test OK {}/{}".format(rhat_count,rhat_count_threshold))
                    elif rhat_count >= rhat_count_threshold:
                        print('... ==> convergence reached')
                        konvergence = k
                        stop = konvergence + supp                       
                #elif (rhat <= rhat_threshold).all() and rhat_count >= rhat_count_threshold:
                #    print '... ==> convergence reached'
                #    konvergence = k
                #    stop = konvergence + supp
                else:
                    rhat_count = 0

        if (k+1) >= stop: #Then we have reached the maximum number of steps for our Markov chain.
            print('We break the loop because we have reached convergence')
            break
      
    if k == nIterations-1:
        print("We have reached the limit number of steps without having converged")
            
    # #########################################################################
    # Construction of the independent samples
    # ######################################################################### 
            
    temp = np.where(chain[0,:,0] == 0.0)[0]
    if len(temp) != 0:
        idxzero = temp[0]
    else:
        idxzero = chain.shape[1]
    
    idx = np.amin([np.floor(2e05/nwalkers),np.floor(0.1*idxzero)])
    if idx == 0:
        isamples = chain[:,0:idxzero,:] 
    else:
        isamples = chain[:,idxzero-idx:idxzero,:]

    if save:
        import pickle
        
        frame = inspect.currentframe()
        args, _, _, values = inspect.getargvalues(frame)
        input_parameters = {j : values[j] for j in args[1:]}        
        
        output = {'isamples':isamples,
                  'chain': chain_zero_truncated(chain),
                  'input_parameters': input_parameters,
                  'AR': sampler.acceptance_fraction,
                  'lnprobability': sampler.lnprobability}
                  
        with open('results/'+output_file+'/MCMC_results','wb') as fileSave:
            myPickler = pickle.Pickler(fileSave)
            myPickler.dump(output)
        
        print('')        
        print("The file MCMC_results has been stored in the folder {}".format('results/'+output_file+'/'))

    if verbose:
        timing(start_time)
                                    
    return chain_zero_truncated(chain)    

                                    
def chain_zero_truncated(chain): 
    """
    Return the Markov chain with the dimension: walkers x steps* x parameters,
    where steps* is the last step before having 0 (not yet constructed chain).
    
    Parameters
    ----------
    chain: numpy.array
        The MCMC chain.
        
    Returns
    -------
    out: numpy.array
        The truncated MCMC chain, that is to say, the chain which only contains 
        relevant information.
    
    """
    try:
        idxzero = np.where(chain[0,:,0] == 0.0)[0][0]
    except:
        idxzero = chain.shape[1]        

    return chain[:,0:idxzero,:]
 
   
def show_walk_plot(chain, save=False, **kwargs):
    """
    Display or save a figure showing the path of each walker during the MCMC run
    
    Parameters
    ----------    
    chain: numpy.array
        The Markov chain. The shape of chain must be nwalkers x length x dim. 
        If a part of the chain is filled with zero values, the method will 
        discard these steps.
    save: boolean, default: False
        If True, a pdf file is created.
    kwargs:
        Additional attributes are passed to the matplotlib plot method.
                                                        
    Returns
    -------
    Display the figure or create a pdf file named walk_plot.pdf in the working
    directory.         
    
    """
    temp = np.where(chain[0,:,0] == 0.0)[0]
    if len(temp) != 0:
        chain = chain[:,:temp[0],:]

    labels = kwargs.pop('labels',["$r$",r"$\theta$","$f$"])
    fig, axes = plt.subplots(3, 1, sharex=True, figsize=kwargs.pop('figsize',(8,6)))
    axes[2].set_xlabel(kwargs.pop('xlabel','step number'))
    axes[2].set_xlim(kwargs.pop('xlim',[0,chain.shape[1]]))
    color = kwargs.pop('color','k')    
    alpha = kwargs.pop('alpha',0.4)
    for j in range(3):            
        axes[j].plot(chain[:,:,j].T, color=color, 
                                     alpha=alpha,
                                     **kwargs)
        axes[j].yaxis.set_major_locator(MaxNLocator(5))
        axes[j].set_ylabel(labels[j])
    fig.tight_layout(h_pad=0.0)
    if save:
        plt.savefig('walk_plot.pdf')
        plt.close(fig)
    else:
        plt.show()                                  


def show_corner_plot(chain, burnin=0.5, save=False, **kwargs):
    """
    Display or save a figure showing the corner plot (pdfs + correlation plots)
    
    Parameters
    ----------
    chain: numpy.array
        The Markov chain. The shape of chain must be nwalkers x length x dim. 
        If a part of the chain is filled with zero values, the method will 
        discard these steps.
    burnin: float, default: 0
        The fraction of a walker we want to discard.
    save: boolean, default: False
        If True, a pdf file is created. 
     
     kwargs:
        Additional attributs are passed to the corner.corner() method.                               
                    
    Returns
    -------
    Display the figure or create a pdf file named walk_plot.pdf in the working
    directory.         
        
    Raises
    ------
    ImportError
    
    """
    try:
        temp = np.where(chain[0,:,0] == 0.0)[0]
        if len(temp) != 0:
            chain = chain[:,:temp[0],:]
        length = chain.shape[1] 
        chain = chain[:,int(np.floor(burnin*(length-1))):length,:].reshape((-1,3))
    except IndexError:
        pass

    if chain.shape[0] == 0:
        print("It seems that the chain is empty. Have you already run the MCMC ?")
    else: 
        fig = corner.corner(chain, labels=kwargs.pop('labels',["$r$",r"$\theta$","$f$"]), **kwargs)
    
    if save:
        plt.savefig('corner_plot.pdf')
        plt.close(fig)
    else:
        plt.show()


def writeText(document,text):
    """
    Write a line of text in a txt file.
    
    Parameters
    ----------
    document: str
        The path to the file to append or create.        
    text: str
        The text to write.
        
    Returns
    -------
    None
        
    """
    with open(document,'a') as fileObject:
        if isinstance(text,str):
            fileObject.write("%s \n" % text)
        elif isinstance(text,tuple):
            defFormat = "%s"
            for k in range(1,len(text)):
                defFormat += "\t %s"
            fileObject.write(defFormat % text)


def confidence(isamples, cfd=68.27, bins=100, gaussianFit=False, weights=None,
               verbose=True, save=False, **kwargs):
    """
    Determine the highly probable value for each model parameter, as well as 
    the 1-sigma confidence interval.
    
    Parameters
    ----------
    isamples: numpy.array
        The independent samples for each model parameter.        
    cfd: float, optional
        The confidence level given in percentage.    
    bins: int, optional
        The number of bins used to sample the posterior distributions.        
    gaussianFit: boolean, optional
        If True, a gaussian fit is performed in order to determine (\mu,\sigma)
    weights : (n, ) array_like or None, optional
        An array of weights for each sample.
    verbose: boolean, optional
        Display information in the shell.
    save: boolean, optional
        If "True", a txt file with the results is saved in the output repository.        
    kwargs: optional
        Additional attributes are passed to the matplotlib hist() method.
        
    Returns
    -------
    out: tuple
        A 2 elements tuple with the highly probable solution and the confidence
        interval.
        
    """
    plsc = kwargs.pop('plsc',0.001)
    title = kwargs.pop('title',None)        
        
    output_file = kwargs.pop('filename','confidence.txt')
        
    try:
        l = isamples.shape[1]        
    except Exception:
        l = 1
     
    confidenceInterval = dict()  
    val_max = dict()
    pKey = ['r','theta','f']
    
    if cfd == 100:
        cfd = 99.9
        
    #########################################    
    ##  Determine the confidence interval  ##
    #########################################
    if gaussianFit:
        mu = np.zeros(3)
        sigma = np.zeros_like(mu)
    
    if gaussianFit:
        fig,ax = plt.subplots(2,3, figsize=(12,8))
    else:
        fig,ax = plt.subplots(1,3, figsize=(12,4))
    
    for j in range(l):               
        label_file = ['r','theta','flux']    
        label = [r'$\Delta r$',r'$\Delta \theta$',r'$\Delta f$']
        
        if gaussianFit:
            n, bin_vertices, _ = ax[0][j].hist(isamples[:,j],bins=bins, 
                                               weights=weights, histtype='step', 
                                               edgecolor='gray')
        else:
            n, bin_vertices, _ = ax[j].hist(isamples[:,j],bins=bins, 
                                            weights=weights, histtype='step', 
                                            edgecolor='gray')
        bins_width = np.mean(np.diff(bin_vertices))
        surface_total = np.sum(np.ones_like(n)*bins_width * n)
        n_arg_sort = np.argsort(n)[::-1]
        
        test = 0
        pourcentage = 0
        for k,jj in enumerate(n_arg_sort):
            test = test + bins_width*n[jj]
            pourcentage = test/surface_total*100.
            if pourcentage > cfd:
                if verbose:
                    print('percentage for {}: {}%'.format(label_file[j],pourcentage))
                break
        n_arg_min = n_arg_sort[:k].min()
        n_arg_max = n_arg_sort[:k+1].max()
        
        if n_arg_min == 0:  n_arg_min += 1
        if n_arg_max == bins:  n_arg_max -= 1        
        
        val_max[pKey[j]] = bin_vertices[n_arg_sort[0]]+bins_width/2.
        confidenceInterval[pKey[j]] = np.array([bin_vertices[n_arg_min-1],
                                                bin_vertices[n_arg_max+1]]-val_max[pKey[j]])
                        
        arg = (isamples[:,j]>=bin_vertices[n_arg_min-1])*(isamples[:,j]<=bin_vertices[n_arg_max+1])
        if gaussianFit:
            _ = ax[0][j].hist(isamples[arg,j],bins=bin_vertices, 
                              facecolor='gray', edgecolor='darkgray', 
                              histtype='stepfilled', alpha=0.5)
            ax[0][j].vlines(val_max[pKey[j]], 0, n[n_arg_sort[0]], 
                            linestyles='dashed', color='red')
            ax[0][j].set_xlabel(label[j])
            if j==0:  ax[0][j].set_ylabel('Counts')
        else:
            _ = ax[j].hist(isamples[arg,j],bins=bin_vertices, facecolor='gray', 
                           edgecolor='darkgray', histtype='stepfilled',
                           alpha=0.5)
            ax[j].vlines(val_max[pKey[j]], 0, n[n_arg_sort[0]],
                         linestyles='dashed', color='red')
            ax[j].set_xlabel(label[j])
            if j==0:  ax[j].set_ylabel('Counts')
    
        if gaussianFit:
            (mu[j], sigma[j]) = norm.fit(isamples[:,j])
            n_fit, bins_fit = np.histogram(isamples[:,j], bins, normed=1, 
                                           weights=weights)
            _= ax[1][j].hist(isamples[:,j], bins, normed=1, weights=weights, 
                             facecolor='gray', edgecolor='darkgray', 
                             histtype='step')
            y = normpdf( bins_fit, mu[j], sigma[j])
            ax[1][j].plot(bins_fit, y, 'r--', linewidth=2, alpha=0.7) 
            
            ax[1][j].set_xlabel(label[j])
            if j==0:  ax[1][j].set_ylabel('Counts')
            
            if title is not None:
                msg = r"$\mu$ = {:.4f}, $\sigma$ = {:.4f}"
                ax[1][j].set_title(title+'   '+msg.format(mu[j],sigma[j]),
                          fontsize=10)
        else:
            if title is not None:            
                ax[1].set_title(title, fontsize=10)

    if save:
        if gaussianFit:
            plt.savefig('confi_hist_flux_r_theta_gaussfit.pdf')
        else:
            plt.savefig('confi_hist_flux_r_theta.pdf')
        
        plt.tight_layout(w_pad=0.001)
        
    if verbose:
        print('')
        print('Confidence intervals:')
        print('r: {} [{},{}]'.format(val_max['r'],
                                     confidenceInterval['r'][0],
                                     confidenceInterval['r'][1]))
        print('theta: {} [{},{}]'.format(val_max['theta'],
                                         confidenceInterval['theta'][0],
                                         confidenceInterval['theta'][1]))
        print('flux: {} [{},{}]'.format(val_max['f'],
                                        confidenceInterval['f'][0],
                                        confidenceInterval['f'][1]))
        if gaussianFit:
            print('')
            print('Gaussian fit results:')
            print('r: {} +-{}'.format(mu[0],sigma[0]))
            print('theta: {} +-{}'.format(mu[1],sigma[1]))
            print('f: {} +-{}'.format(mu[2],sigma[2]))

    ##############################################
    ##  Write inference results in a text file  ##
    ##############################################    
    if save:         
        try:
            fileObject = open(output_file,'r')
        except IOError: # if the file doesn't exist, we create it (empty)
            answer = 'y'
            if answer == 'y':
                fileObject = open(output_file,'w')
            elif answer == 'n':
                msg = "The file has not been created. The object cannot be "
                msg += "created neither."
                print(msg)
                raise IOError("No such file has been found")
            else:
                msg = "You must choose between 'y' for yes and 'n' for no. The "
                msg += "file has not been created. The object cannot be "
                msg += "created neither."
                print()
                raise IOError("No such file has been found")
        finally:
            fileObject.close()    
    
        writeText(output_file,'###########################')
        writeText(output_file,'####   INFERENCE TEST   ###')
        writeText(output_file,'###########################')
        writeText(output_file,' ')
        writeText(output_file,'Results of the MCMC fit')
        writeText(output_file,'----------------------- ')
        writeText(output_file,' ')
        writeText(output_file,'>> Position and flux of the planet (highly probable):')
        writeText(output_file,'{} % confidence interval'.format(cfd))
        writeText(output_file,' ')
        for i in range(3):
            confidenceMax = confidenceInterval[pKey[i]][1]
            confidenceMin = -confidenceInterval[pKey[i]][0]
            if i == 2:
                text = '{}: \t\t\t{:.3f} \t-{:.3f} \t+{:.3f}'
            else:
                text = '{}: \t\t\t{:.3f} \t\t-{:.3f} \t\t+{:.3f}'
                
            writeText(output_file,text.format(pKey[i],val_max[pKey[i]],
                                              confidenceMin,confidenceMax))                   
        
        writeText(output_file,' ')
        writeText(output_file,'Platescale = {} mas'.format(plsc*1000))
        text = '{}: \t\t{:.2f} \t\t-{:.2f} \t\t+{:.2f}'
        writeText(output_file,text.format('r (mas)', val_max[pKey[0]]*plsc*1000,
                                          -confidenceInterval[pKey[0]][0]*plsc*1000,
                                          confidenceInterval[pKey[0]][1]*plsc*1000))

    if gaussianFit:
        return (mu,sigma)
    else:
        return (val_max,confidenceInterval)           

        


#! /usr/bin/env python

"""
Module with frame de-rotation routine for ADI.
"""
__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['cube_derotate',
           'frame_rotate']

import numpy as np
import warnings
try:
    import cv2
    no_opencv = False
except ImportError:
    msg = "Opencv python bindings are missing."
    warnings.warn(msg, ImportWarning)
    no_opencv = True

from skimage.transform import rotate
from multiprocessing import Pool, cpu_count
import itertools as itt
from ..conf.utils_conf import eval_func_tuple as futup
from ..var import frame_center

data_array = None # holds the (implicitly mem-shared) data array


def frame_rotate(array, angle, imlib='opencv', interpolation='lanczos4',
                 cxy=None):
    """ Rotates a frame or 2D array.
    
    Parameters
    ----------
    array : array_like 
        Input image, 2d array.
    angle : float
        Rotation angle.
    imlib : {'opencv', 'skimage'}, str optional
        Library used for image transformations. Opencv is faster than ndimage or
        skimage.
    interpolation : str, optional
        For 'skimage' library: 'nearneig', bilinear', 'bicuadratic', 'bicubic',
        'biquartic', 'biquintic'. The 'nearneig' interpolation is the fastest
        and the 'biquintic' the slowest. The 'nearneig' is the poorer
        option for interpolation of noisy astronomical images.
        For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
        The 'nearneig' interpolation is the fastest and the 'lanczos4' the
        slowest and accurate. 'lanczos4' is the default.
    cxy : float, optional
        Coordinates X,Y  of the point with respect to which the rotation will be 
        performed. By default the rotation is done with respect to the center 
        of the frame; central pixel if frame has odd size.
        
    Returns
    -------
    array_out : array_like
        Resulting frame.
        
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')

    y, x = array.shape
    
    if cxy is None:
        cy, cx = frame_center(array)
    else:
        cx, cy = cxy

    if imlib == 'skimage':
        if interpolation == 'nearneig':
            order = 0
        elif interpolation == 'bilinear':
            order = 1
        elif interpolation == 'bicuadratic':
            order = 2
        elif interpolation == 'bicubic':
            order = 3
        elif interpolation == 'biquartic':
            order = 4
        elif interpolation == 'biquintic':
            order = 5
        else:
            raise TypeError('Skimage interpolation method not recognized.')

        min_val = np.min(array)
        im_temp = array - min_val
        max_val = np.max(im_temp)
        im_temp /= max_val

        array_out = rotate(im_temp, angle, order=order, center=cxy, cval=np.nan)

        array_out *= max_val
        array_out += min_val
        array_out = np.nan_to_num(array_out)

    elif imlib == 'opencv':
        if no_opencv:
            msg = 'Opencv python bindings cannot be imported. Install opencv or'
            msg += ' set imlib to skimage.'
            raise RuntimeError(msg)

        if interpolation == 'bilinear':
            intp = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            intp= cv2.INTER_CUBIC
        elif interpolation == 'nearneig':
            intp = cv2.INTER_NEAREST
        elif interpolation == 'lanczos4':
            intp = cv2.INTER_LANCZOS4
        else:
            raise TypeError('Opencv interpolation method not recognized.')

        M = cv2.getRotationMatrix2D((cx,cy), angle, 1)
        array_out = cv2.warpAffine(array.astype(np.float32), M, (x, y),
                                   flags=intp)

    else:
        raise ValueError('Image transformation library not recognized.')
             
    return array_out
    
    
def cube_derotate(array, angle_list, imlib='opencv', interpolation='lanczos4',
                  cxy=None, nproc=1):
    """ Rotates an cube (3d array or image sequence) providing a vector or
    corrsponding angles. Serves for rotating an ADI sequence to a common north
    given a vector with the corresponding parallactic angles for each frame. By
    default bicubic interpolation is used (opencv).
    
    Parameters
    ----------
    array : array_like 
        Input 3d array, cube.
    angle_list : list
        Vector containing the parallactic angles.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    cxy : tuple of int, optional
        Coordinates X,Y  of the point with respect to which the rotation will be 
        performed. By default the rotation is done with respect to the center 
        of the frames, as it is returned by the function vip_hci.var.frame_center. 
    collapse : {'median','mean'}
        Way of collapsing the derotated cube.
        
    Returns
    -------
    array_der : array_like
        Resulting cube with de-rotated frames.
        
    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array.')
    array_der = np.zeros_like(array)
    n_frames = array.shape[0]

    if not nproc:
        nproc = int((cpu_count() / 2))

    if nproc == 1:
        for i in range(n_frames):
            array_der[i] = frame_rotate(array[i], -angle_list[i], imlib=imlib,
                                        interpolation=interpolation, cxy=cxy)
    elif nproc > 1:
        global data_array
        data_array = array

        pool = Pool(processes=int(nproc))
        res = pool.map(futup, zip(itt.repeat(_cube_rotate_mp),
                                  range(n_frames), itt.repeat(angle_list),
                                  itt.repeat(imlib), itt.repeat(interpolation),
                                  itt.repeat(cxy)))
        pool.close()
        array_der = np.array(res)

    return array_der


def _cube_rotate_mp(num_fr, angle_list, imlib, interpolation, cxy):
    framerot = frame_rotate(data_array[num_fr], -angle_list[num_fr],
                            imlib, interpolation, cxy)
    return framerot



#! /usr/bin/env python

"""
Module with various fits handling functions.
"""
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['open_fits',
           'open_adicube',
           'info_fits',
           'write_fits',
           'append_extension',
           'verify_fits',
           'byteswap_array']


import os
import numpy as np
from astropy.io import fits as ap_fits


def open_fits(fitsfilename, n=0, header=False, ignore_missing_end=False,
              precision=np.float32, verbose=True):
    """Loads a fits file into a memory as numpy array.
    
    Parameters
    ----------
    fitsfilename : string
        Name of the fits file.
    n : int
        It chooses which HDU to open. Default is the first one.
    header : {False, True}, bool optional
        Whether to return the header along with the data or not.
    precision : numpy dtype
        Float precision, by default np.float32 or single precision float.
    ignore_missing_end : {False, True}, bool optional
        Allows to open fits files with a header missing END card.
    verbose : {True, False}, bool optional
        If True prints message of completion.
    
    Returns
    -------
    data : array_like
        Array containing the frames of the fits-cube.
    If header is True:
    header : dictionary
        Dictionary containing the fits header.
    """
    if not fitsfilename.endswith('.fits'):
        fitsfilename = str(fitsfilename+'.fits')
    hdulist = ap_fits.open(fitsfilename, memmap=True,
                           ignore_missing_end=ignore_missing_end)
    data = hdulist[n].data
    data = np.array(data, dtype=precision)
    
    if header:
        header = hdulist[0].header
        if verbose:
            if len(data.shape)==1:
                msg = "\nFits HDU:{:} data and header successfully loaded. Data"
                msg += " shape: [{:}]"
                print(msg.format(n, data.shape[0]))
            if len(data.shape)==2:  
                msg = "\nFits HDU:{:} data and header successfully loaded. Data"
                msg += " shape: [{:},{:}]"
                print(msg.format(n, data.shape[0],data.shape[1]))
            if len(data.shape)==3:
                msg = "\nFits HDU:{:} data and header successfully loaded. Data"
                msg += " shape: [{:},{:},{:}]"
                print(msg.format(n, data.shape[0],data.shape[1],data.shape[2]))
            if len(data.shape)==4:
                msg = "\nFits HDU:{:} data and header successfully loaded. Data"
                msg += " shape: [{:},{:},{:},{:}]"
                print(msg.format(n, data.shape[0], data.shape[1], data.shape[2],
                                 data.shape[3]))
        hdulist.close()
        return data, header 
    else:
        if verbose:
            if len(data.shape)==1:
                msg = "\nFits HDU:{:} data successfully loaded. Data"
                msg += " shape: [{:}]"
                print(msg.format(n, data.shape[0]))
            if len(data.shape)==2:  
                msg = "\nFits HDU:{:} data successfully loaded. Data"
                msg += " shape: [{:},{:}]"
                print(msg.format(n, data.shape[0],data.shape[1]))
            if len(data.shape)==3:
                msg = "\nFits HDU:{:} data successfully loaded. Data"
                msg += " shape: [{:},{:},{:}]"
                print(msg.format(n, data.shape[0],data.shape[1],data.shape[2]))
            if len(data.shape)==4:
                msg = "\nFits HDU:{:} data successfully loaded. Data"
                msg += " shape: [{:},{:},{:},{:}]"
                print(msg.format(n, data.shape[0], data.shape[1], data.shape[2],
                                 data.shape[3]))
        hdulist.close()
        return data        


def open_adicube(fitsfilename, verbose=True):
    """ Opens an ADI cube with the parallactic angles appended (see function
    append_par_angles).
    
    Parameters
    ----------
    fitsfilename : string
        Name of the fits file.
    verbose : {True, False}, bool optional
        If True prints message.
        
    Returns
    -------
    data : array_like
        Array containing the frames of the fits-cube.
    parangles : array_like
        1d array containing the corresponding parallactic angles.
          
    """
    if not fitsfilename.endswith('.fits'):
        fitsfilename = str(fitsfilename+'.fits')
    hdulist = ap_fits.open(fitsfilename, memmap=True)
    data = hdulist[0].data
    if not data.ndim ==3:
        raise TypeError('Input fits file does not contain a cube or 3d array.')
    parangles = hdulist[1].data 
    if verbose:
        msg1 = "\nFits HDU:{:} data successfully loaded. Data shape: [{:},{:},{:}]"
        msg2 = "\nFits HDU:{:} data successfully loaded. Data shape: [{:}]"
        print(msg1.format(0, data.shape[0],data.shape[1],data.shape[2]))
        print(msg2.format(1, parangles.shape[0]))
    
    return data, parangles


def byteswap_array(array):
    """ FITS files are stored in big-endian byte order. All modern CPUs are 
    little-endian byte order, so at some point you have to byteswap the data. 
    Some FITS readers (cfitsio, the fitsio python module) do the byteswap when 
    reading the data from disk to memory, so we get numpy arrays in native 
    (little-endian) byte order. Unfortunately, astropy.io.fits does not 
    byteswap for us, and we get numpy arrays in non-native byte order. 
    However, most of the time we never notice this because when you do any 
    numpy operations on such arrays, numpy uses an intermediate buffer to 
    byteswap the array behind the scenes and returns the result as a native 
    byte order array. Some operations require the data to be byteswaped 
    before and will complain about it. This function will help in this cases.
    
    Parameters
    ----------
    array : array_like
        2d input array.
        
    Returns
    -------
    array_out : array_like
        2d resulting array after the byteswap operation.


    Notes
    -----
    http://docs.scipy.org/doc/numpy-1.10.1/user/basics.byteswapping.html

    """
    array_out = array.byteswap().newbyteorder()
    return array_out


def info_fits(fitsfilename):
    """Prints the information about a fits file. 
    """
    hdulist = ap_fits.open(fitsfilename, memmap=True)
    hdulist.info()

         
def verify_fits(fitspath):
    """Verifies "the FITS standard" of a fits file or list of fits.

    Parameters
    ----------
    fitspath : string
        Path to the fits file or list with fits filename paths.

    """
    if isinstance(fitspath, list):
        for ffile in fitspath:
            f = ap_fits.open(ffile)
            f.verify()
    else:
        f = ap_fits.open(fitspath)
        f.verify()
    
    
def write_fits(filename, array, header=None, dtype32=True, verbose=True):
    """Writes array and header into FTIS file, if there is a previous file with
    the same filename then it's replaced.
    
    Parameters
    ----------
    filename : string
        Filename of the fits file to be written.
    array : array_like
        Array to be written into a fits file.
    header : array_like, optional
        Array with header. 
    dtype32 : {True, False}, bool optional
        If True the array is casted as a float32
    verbose : {True, False}, bool optional
        If True prints message.

    """
    if dtype32:
        array = array.astype('float32', copy=False)
    if os.path.exists(filename):
        os.remove(filename)                                     
        if verbose:
            print("\nFits file successfully overwritten")
        ap_fits.writeto(filename, array, header)
    else:
        ap_fits.writeto(filename, array, header)
        if verbose:
            print("\nFits file successfully saved")    


def append_extension(filename, array):
    """Appends an extension to fits file. 
    """
    ap_fits.append(filename, array)
    print("\nFits extension appended")
        
        
    


"""
Subpackage ``phot`` includes functionalities such as:
    - signal-to-noise (S/R) estimation,
    - S/R map generation,
    - detection of point like sources (for pipelines),
    - fake companions injection,
    - algorithms throughput,
    - contrast curve generation.
"""
from __future__ import absolute_import

from .contrcurve import *
from .detection import *
from .fakecomp import *
from .frame_analysis import *
from .snr import *



#! /usr/bin/env python

"""
Module with local smart pca (annulus-wise) serial and parallel implementations.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['pca_adi_annular',
           'pca_rdi_annular']

import numpy as np
import itertools as itt
from scipy import stats
from multiprocessing import Pool, cpu_count
from ..preproc import cube_derotate, cube_collapse, check_pa_vector
from ..conf import time_ini, timing
from ..conf.utils_conf import eval_func_tuple as EFT
from ..var import get_annulus_segments, matrix_scaling, get_annulus
from ..madi.adi_utils import _find_indices, _define_annuli
from ..stats import descriptive_stats
from .svd import get_eigenvectors


def pca_rdi_annular(cube, angle_list, cube_ref, radius_int=0, asize=1, 
                    ncomp=1, svd_mode='randsvd', min_corr=0.9, fwhm=4, 
                    scaling='temp-standard', imlib='opencv',
                    interpolation='lanczos4', collapse='median',
                    full_output=False, verbose=True):
    """ Annular PCA with Reference Library + Correlation + standardization
    
    In the case of having a large number of reference images, e.g. for a survey 
    on a single instrument, we can afford a better selection of the library by 
    constraining the correlation with the median of the science dataset and by 
    working on an annulus-wise way. As with other local PCA algorithms in VIP
    the number of principal components can be automatically adjusted by the
    algorithm by minmizing the residuals in the given patch (a la LOCI). 
    
    Parameters
    ----------
    cube : array_like, 3d
        Input science cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    cube_ref : array_like, 3d
        Reference library cube. For Reference Star Differential Imaging.
    radius_int : int, optional
        The radius of the innermost annulus. By default is 0, if >0 then the 
        central circular area is discarded.
    asize : int, optional
        The size of the annuli, in FWHM. Default is 3.
    ncomp : int, optional
        How many PCs are kept. If none it will be automatically determined.
    svd_mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy', 'randcupy'}, str
        Switch for the SVD method/library to be used. ``lapack`` uses the LAPACK 
        linear algebra library through Numpy and it is the most conventional way 
        of computing the SVD (deterministic result computed on CPU). ``arpack`` 
        uses the ARPACK Fortran libraries accessible through Scipy (computation
        on CPU). ``eigen`` computes the singular vectors through the 
        eigendecomposition of the covariance M.M' (computation on CPU).
        ``randsvd`` uses the randomized_svd algorithm implemented in Sklearn 
        (computation on CPU). ``cupy`` uses the Cupy library for GPU computation
        of the SVD as in the LAPACK version. ``eigencupy`` offers the same 
        method as with the ``eigen`` option but on GPU (through Cupy). 
        ``randcupy`` is an adaptation of the randomized_svd algorith, where all 
        the computations are done on a GPU. 
    min_corr : int, optional
        Level of linear correlation between the library patches and the median 
        of the science. Deafult is 0.9.
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Deafult is 4.
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    full_output: boolean, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
    verbose : {True, False}, bool optional
        If True prints to stdout intermediate info.
    
    Returns
    -------
    frame : array_like, 2d    
        Median combination of the de-rotated cube.
    If full_output is True:  
    array_out : array_like, 3d 
        Cube of residuals.
    array_der : array_like, 3d
        Cube residuals after de-rotation.    
    
    """
    def define_annuli(angle_list, ann, n_annuli, fwhm, radius_int, annulus_width, 
                      verbose):
        """ Defining the annuli """
        if ann == n_annuli-1:
            inner_radius = radius_int + (ann*annulus_width-1)
        else:                                                                                         
            inner_radius = radius_int + ann*annulus_width
        ann_center = (inner_radius+(annulus_width/2.0))
        
        if verbose:
            msg2 = 'Annulus {:}, Inn radius = {:.2f}, Ann center = {:.2f} '
            print(msg2.format(int(ann+1),inner_radius, ann_center))
        return inner_radius, ann_center
    
    def fr_ref_correlation(vector, matrix):
        """ Getting the correlations """
        lista = []
        for i in range(matrix.shape[0]):
            pears, _ = stats.pearsonr(vector, matrix[i])
            lista.append(pears)
        
        return lista

    def do_pca_annulus(ncomp, matrix, svd_mode, noise_error, data_ref):
        """ Actual PCA for the annulus """
        #V = svd_wrapper(data_ref, svd_mode, ncomp, debug=False, verbose=False)
        V = get_eigenvectors(ncomp, matrix, svd_mode, noise_error=noise_error, 
                             data_ref=data_ref, debug=False)
        # new variables as linear combinations of the original variables in 
        # matrix.T with coefficientes from EV
        transformed = np.dot(V, matrix.T) 
        reconstructed = np.dot(V.T, transformed)
        residuals = matrix - reconstructed.T  
        return residuals, V.shape[0]

    #---------------------------------------------------------------------------
    array = cube
    array_ref = cube_ref
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array.')
    if not array.shape[0] == angle_list.shape[0]:
        raise TypeError('Input vector or parallactic angles has wrong length.')
    
    n, y, _ = array.shape
    if verbose:  start_time = time_ini()
    
    angle_list = check_pa_vector(angle_list)

    annulus_width = asize * fwhm                     # equal size for all annuli
    n_annuli = int(np.floor((y/2-radius_int)/annulus_width))
    if verbose:
        msg = '# annuli = {:}, Ann width = {:}, FWHM = {:.3f}\n'
        print(msg.format(n_annuli, annulus_width, fwhm))
        print('PCA will be done locally per annulus and per quadrant.\n')
     
    cube_out = np.zeros_like(array)
    for ann in range(n_annuli):
        inner_radius, _ = define_annuli(angle_list, ann, n_annuli, fwhm, 
                                        radius_int, annulus_width, verbose) 
        indices = get_annulus(array[0], inner_radius, annulus_width,
                              output_indices=True)
        yy = indices[0]
        xx = indices[1]
                    
        matrix = array[:, yy, xx]                 # shape [nframes x npx_ann] 
        matrix_ref = array_ref[:, yy, xx]
        
        corr = fr_ref_correlation(np.median(matrix, axis=0), matrix_ref)
        indcorr = np.where(np.abs(corr)>=min_corr)
        #print indcorr
        data_ref = matrix_ref[indcorr]
        nfrslib = data_ref.shape[0]
                
        if nfrslib<5:
            msg = 'Too few frames left (<5) fulfill the given correlation level.'
            msg += 'Try decreasing it'
            raise RuntimeError(msg)
        
        matrix = matrix_scaling(matrix, scaling)
        data_ref = matrix_scaling(data_ref, scaling)
        
        residuals, ncomps = do_pca_annulus(ncomp, matrix, svd_mode, 10e-3, data_ref)  
        cube_out[:, yy, xx] = residuals  
            
        if verbose:
            print('# frames in LIB = {}'.format(nfrslib))
            print('# PCs = {}'.format(ncomps))
            print('Done PCA with {:} for current annulus'.format(svd_mode))
            timing(start_time)      
         
    cube_der = cube_derotate(cube_out, angle_list, imlib=imlib,
                             interpolation=interpolation)
    frame = cube_collapse(cube_der, mode=collapse)
    if verbose:
        print('Done derotating and combining.')
        timing(start_time)
    if full_output:
        return cube_out, cube_der, frame 
    else:
        return frame           


def pca_adi_annular(cube, angle_list, radius_int=0, fwhm=4, asize=3,
                    n_segments=1, delta_rot=1, ncomp=1, svd_mode='randsvd',
                    nproc=1, min_frames_lib=10, max_frames_lib=200, tol=1e-1,
                    scaling=None, imlib='opencv', interpolation='lanczos4',
                    collapse='median', full_output=False, verbose=True):
    """ Annular (smart) ADI PCA. The PCA model is computed locally in each
    annulus (optionally quadrants of each annulus). For each annulus we discard
    reference images taking into account a parallactic angle threshold
    (set by ``delta_rot``).
     
    Depending on parameter ``nproc`` the algorithm can work with several cores.
    It's been tested on a Linux and OSX. The ACCELERATE library for linear 
    algebra calcularions, which comes by default in every OSX system, is broken 
    for multiprocessing. Avoid using this function unless you have compiled 
    Python against other linear algebra library. An easy fix is to install 
    (ana)conda and the openblas or MKL libraries (replacing ACCELERATE).
    
    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    radius_int : int, optional
        The radius of the innermost annulus. By default is 0, if >0 then the 
        central circular area is discarded.
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Deafult is 4.
    asize : int, optional
        The size of the annuli, in FWHM. Default is 3.
    n_segments : int or list of ints, optional
        The number of segments for each annulus. When a single integer is given
        it is used for all annuli.
    delta_rot : int, optional
        Factor for increasing the parallactic angle threshold, expressed in FWHM.
        Default is 1 (excludes 1 FHWM on each side of the considered frame).
        According to Absil+13, a slightly better contrast can be reached for the 
        innermost annuli if we consider a ``delta_rot`` condition as small as 
        0.1 lambda/D. This is because at very small separation, the effect of 
        speckle correlation is more significant than self-subtraction.
    ncomp : int or list or 1d numpy array, optional
        How many PCs are kept. If none it will be automatically determined. If a
        list is provided and it matches the number of annuli then a different
        number of PCs will be used for each annulus (starting with the innermost
        one).
    svd_mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy', 'randcupy'}, str
        Switch for the SVD method/library to be used. ``lapack`` uses the LAPACK 
        linear algebra library through Numpy and it is the most conventional way 
        of computing the SVD (deterministic result computed on CPU). ``arpack`` 
        uses the ARPACK Fortran libraries accessible through Scipy (computation
        on CPU). ``eigen`` computes the singular vectors through the 
        eigendecomposition of the covariance M.M' (computation on CPU).
        ``randsvd`` uses the randomized_svd algorithm implemented in Sklearn 
        (computation on CPU). ``cupy`` uses the Cupy library for GPU computation
        of the SVD as in the LAPACK version. ``eigencupy`` offers the same 
        method as with the ``eigen`` option but on GPU (through Cupy). 
        ``randcupy`` is an adaptation of the randomized_svd algorith, where all 
        the computations are done on a GPU. 
    nproc : None or int, optional
        Number of processes for parallel computing. If None the number of 
        processes will be set to (cpu_count()/2). By default the algorithm works
        in single-process mode.
    min_frames_lib : int, optional
        Minimum number of frames in the PCA reference library. Be careful, when
        ``min_frames_pca`` < ``ncomp``, then for certain frames the subtracted
        low-rank approximation is not possible. It is recommended to decrease
        ``delta_rot`` and have enough frames in the libraries to allow getting 
        ``ncomp`` PCs.
    max_frames_lib : int, optional
        Maximum number of frames in the PCA reference library for annuli beyond
        10*FWHM. The more distant/decorrelated frames are removed from the
        library.
    tol : float, optional
        Stopping criterion for choosing the number of PCs when ``ncomp`` is None.
        Lower values will lead to smaller residuals and more PCs.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    full_output: boolean, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.  
    verbose : {True, False}, bool optional
        If True prints to stdout intermediate info.
     
    Returns
    -------
    frame : array_like, 2d    
        Median combination of the de-rotated cube.
    If full_output is True:  
    array_out : array_like, 3d 
        Cube of residuals.
    array_der : array_like, 3d
        Cube residuals after de-rotation.
     
    """
    array = cube
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')
    if not array.shape[0] == angle_list.shape[0]:
        raise TypeError('Input vector or parallactic angles has wrong length')

    n, y, _ = array.shape
     
    if verbose:
        start_time = time_ini()
    
    angle_list = check_pa_vector(angle_list)

    annulus_width = asize * fwhm                     # equal size for all annuli
    n_annuli = int(np.floor((y/2 - radius_int) / annulus_width))

    if isinstance(n_segments, int):
        n_segments = [n_segments for _ in range(n_annuli)]
    elif n_segments == 'auto':
        n_segments = []
        n_segments.append(2)  # for first annulus
        n_segments.append(3)  # for second annulus
        ld = 2 * np.tan(360 / 4 / 2) * annulus_width
        for i in range(2, n_annuli):  # rest of annuli
            radius = i * annulus_width
            ang = np.rad2deg(2 * np.arctan(ld / (2 * radius)))
            n_segments.append(int(np.ceil(360 / ang)))

    if verbose:
        msg = '# annuli = {:}, Ann width = {:}, FWHM = {:.3f} \n'
        print(msg.format(n_annuli, annulus_width, fwhm))
        print('PCA per annulus (or annular segment)\n')

    if nproc is None:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = (cpu_count() / 2)
    
    #***************************************************************************
    # The annuli are built, and the corresponding PA thresholds for frame 
    # rejection are calculated (at the center of the annulus)
    #***************************************************************************
    cube_out = np.zeros_like(array)
    for ann in range(n_annuli):
        if isinstance(ncomp, list) or isinstance(ncomp, np.ndarray):
            ncomp = list(ncomp)
            if len(ncomp) == n_annuli:
                ncompann = ncomp[ann]
            else:
                msge = 'If ncomp is a list, it must match the number of annuli'
                raise TypeError(msge)
        else:
            ncompann = ncomp

        n_segments_ann = n_segments[ann]
        res_ann_par = _define_annuli(angle_list, ann, n_annuli, fwhm,
                                     radius_int, annulus_width, delta_rot,
                                     n_segments_ann, verbose)
        pa_thr, inner_radius, ann_center = res_ann_par
        indices = get_annulus_segments(array[0], inner_radius,
                                       annulus_width, n_segments_ann, 0)

        # **********************************************************************
        # Library matrix is created for each segment and scaled if needed
        # **********************************************************************
        for j in range(n_segments_ann):
            yy = indices[j][0]
            xx = indices[j][1]
            matrix_segm = array[:, yy, xx]  # shape [nframes x npx_segment]
            matrix_segm = matrix_scaling(matrix_segm, scaling)

            # ***************************************************************
            # We loop the frames and do the PCA to obtain the residuals cube
            # ***************************************************************
            residuals = do_pca_loop(matrix_segm, nproc, angle_list, fwhm,
                                    pa_thr, ann_center, svd_mode, ncompann,
                                    min_frames_lib, max_frames_lib, tol,
                                    verbose)
            for frame in range(n):
                cube_out[frame][yy, xx] = residuals[frame]

        if verbose:
            print('Done PCA with {:} for current annulus'.format(svd_mode))
            timing(start_time)      
         
    #***************************************************************************
    # Cube is derotated according to the parallactic angle and median combined.
    #***************************************************************************
    cube_der = cube_derotate(cube_out, angle_list, imlib=imlib,
                             interpolation=interpolation)
    frame = cube_collapse(cube_der, mode=collapse)
    if verbose:
        print('Done derotating and combining.')
        timing(start_time)
    if full_output:
        return cube_out, cube_der, frame 
    else:
        return frame 
        
    
    
################################################################################
### Help functions (encapsulating portions of the main algos)
################################################################################
    
def do_pca_loop(matrix, nproc, angle_list, fwhm, pa_threshold, ann_center,
                svd_mode, ncomp, min_frames_lib, max_frames_lib, tol, verbose):
    """
    """
    matrix_ann = matrix
    n = matrix.shape[0]
    #***************************************************************
    # For each frame we call the subfunction do_pca_patch that will 
    # do PCA on the small matrix, where some frames are discarded 
    # according to the PA threshold, and return the residuals
    #***************************************************************          
    ncomps = []
    nfrslib = []          
    if nproc == 1:
        residualarr = []
        for frame in range(n):
            res = do_pca_patch(matrix_ann, frame, angle_list, fwhm,
                               pa_threshold, ann_center, svd_mode, ncomp,
                               min_frames_lib, max_frames_lib, tol)
            residualarr.append(res[0])
            ncomps.append(res[1])
            nfrslib.append(res[2])
        residuals = np.array(residualarr)
        
    elif nproc > 1:
        #***********************************************************
        # A multiprocessing pool is created to process the frames in 
        # a parallel way. SVD/PCA is done in do_pca_patch function
        #***********************************************************            
        pool = Pool(processes=int(nproc))
        res = pool.map(EFT, zip(itt.repeat(do_pca_patch),
                                itt.repeat(matrix_ann), range(n),
                                itt.repeat(angle_list), itt.repeat(fwhm),
                                itt.repeat(pa_threshold),
                                itt.repeat(ann_center), itt.repeat(svd_mode),
                                itt.repeat(ncomp), itt.repeat(min_frames_lib),
                                itt.repeat(max_frames_lib), itt.repeat(tol)))
        res = np.array(res)
        residuals = np.array(res[:,0])
        ncomps = res[:, 1]
        nfrslib = res[:, 2]
        pool.close()                         

    # number of frames in library printed for each annular quadrant
    if verbose:
        descriptive_stats(nfrslib, verbose=verbose, label='Size LIB: ')
    # number of PCs printed for each annular quadrant     
    if ncomp is None and verbose:  
        descriptive_stats(ncomps, verbose=verbose, label='Numb PCs: ')
        
    return residuals


def do_pca_patch(matrix, frame, angle_list, fwhm, pa_threshold, ann_center,
                 svd_mode, ncomp, min_frames_lib, max_frames_lib,  tol):
    """
    Does the SVD/PCA for each frame patch (small matrix). For each frame we 
    find the frames to be rejected depending on the amount of rotation. The
    library is also truncated on the other end (frames too far or which have 
    rotated more) which are more decorrelated to keep the computational cost 
    lower. This truncation is done on the annuli after 10*FWHM and the goal is
    to keep min(num_frames/2, 200) in the library. 
    """
    if pa_threshold != 0:
        if ann_center > fwhm*10:    # TODO: 10*FWHM optimal? new parameter?
            indices_left = _find_indices(angle_list, frame, pa_threshold,
                                         truncate=True,
                                         max_frames=max_frames_lib)
        else:
            indices_left = _find_indices(angle_list, frame, pa_threshold,
                                         truncate=False)
         
        data_ref = matrix[indices_left]
        
        if data_ref.shape[0] <= min_frames_lib:
            msg = 'Too few frames left in the PCA library. '
            msg += 'Try decreasing either delta_rot or min_frames_lib.'
            raise RuntimeError(msg)
    else:
        data_ref = matrix
       
    data = data_ref
    #data = data_ref - data_ref.mean(axis=0)
    curr_frame = matrix[frame]                     # current frame
    
    V = get_eigenvectors(ncomp, data, svd_mode, noise_error=tol, debug=False)        
    
    transformed = np.dot(curr_frame, V.T)
    reconstructed = np.dot(transformed.T, V)                        
    residuals = curr_frame - reconstructed     
    return residuals, V.shape[0], data_ref.shape[0]  




